################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from pyflink.common import SerializationSchema, DeserializationSchema
from pyflink.datastream.functions import SinkFunction, SourceFunction
from pyflink.java_gateway import get_gateway


__all__ = [
    'RMQConnectionConfig',
    'RMQSource',
    'RMQSink'
]


class RMQConnectionConfig(object):
    """
    Connection Configuration for RMQ.
    """

    def __init__(self, j_rmq_connection_config):
        self._j_rmq_connection_config = j_rmq_connection_config

    def get_host(self) -> str:
        return self._j_rmq_connection_config.getHost()

    def get_port(self) -> int:
        return self._j_rmq_connection_config.getPort()

    def get_virtual_host(self) -> str:
        return self._j_rmq_connection_config.getVirtualHost()

    def get_user_name(self) -> str:
        return self._j_rmq_connection_config.getUsername()

    def get_password(self) -> str:
        return self._j_rmq_connection_config.getPassword()

    def get_uri(self) -> str:
        return self._j_rmq_connection_config.getUri()

    def get_network_recovery_interval(self) -> int:
        return self._j_rmq_connection_config.getNetworkRecoveryInterval()

    def is_automatic_recovery(self) -> bool:
        return self._j_rmq_connection_config.isAutomaticRecovery()

    def is_topology_recovery(self) -> bool:
        return self._j_rmq_connection_config.isTopologyRecovery()

    def get_connection_timeout(self) -> int:
        return self._j_rmq_connection_config.getConnectionTimeout()

    def get_requested_channel_max(self) -> int:
        return self._j_rmq_connection_config.getRequestedChannelMax()

    def get_requested_frame_max(self) -> int:
        return self._j_rmq_connection_config.getRequestedFrameMax()

    def get_requested_heartbeat(self) -> int:
        return self._j_rmq_connection_config.getRequestedHeartbeat()

    class Builder(object):
        """
        Builder for RMQConnectionConfig.
        """

        def __init__(self):
            self._j_options_builder = get_gateway().jvm.org.apache.flink.streaming.connectors\
                .rabbitmq.common.RMQConnectionConfig.Builder()

        def set_port(self, port: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setPort(port)
            return self

        def set_host(self, host: str) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setHost(host)
            return self

        def set_virtual_host(self, vhost: str) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setVirtualHost(vhost)
            return self

        def set_user_name(self, user_name: str) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setUserName(user_name)
            return self

        def set_password(self, password: str) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setPassword(password)
            return self

        def set_uri(self, uri: str) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setUri(uri)
            return self

        def set_topology_recovery_enabled(
                self, topology_recovery_enabled: bool) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setTopologyRecoveryEnabled(topology_recovery_enabled)
            return self

        def set_requested_heartbeat(
                self, requested_heartbeat: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setRequestedHeartbeat(requested_heartbeat)
            return self

        def set_requested_frame_max(
                self, requested_frame_max: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setRequestedFrameMax(requested_frame_max)
            return self

        def set_requested_channel_max(
                self, requested_channel_max: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setRequestedChannelMax(requested_channel_max)
            return self

        def set_network_recovery_interval(
                self, network_recovery_interval: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setNetworkRecoveryInterval(network_recovery_interval)
            return self

        def set_connection_timeout(self, connection_timeout: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setConnectionTimeout(connection_timeout)
            return self

        def set_automatic_recovery(self, automatic_recovery: bool) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setAutomaticRecovery(automatic_recovery)
            return self

        def set_prefetch_count(self, prefetch_count: int) -> 'RMQConnectionConfig.Builder':
            self._j_options_builder.setPrefetchCount(prefetch_count)
            return self

        def build(self) -> 'RMQConnectionConfig':
            return RMQConnectionConfig(self._j_options_builder.build())


class RMQSource(SourceFunction):
    def __init__(self,
                 connection_config: 'RMQConnectionConfig',
                 queue_name: str,
                 use_correlation_id: bool,
                 deserialization_schema: DeserializationSchema
                 ):
        """
        Creates a new RabbitMQ source.

        For exactly-once, you must set the correlation ids of messages at the producer.
        The correlation id must be unique. Otherwise the behavior of the source is undefined.

        If in doubt, set use_correlation_id to False.

        When correlation ids are not used, this source has at-least-once processing semantics
        when checkpointing is enabled.

        :param connection_config: The RabbiMQ connection configuration.
        :param queue_name: The queue to receive messages from.
        :param use_correlation_id: Whether the messages received are supplied with a unique id
                                   to deduplicate messages (in case of failed acknowledgments).
                                   Only used when checkpointing is enabled.
        :param deserialization_schema: A deserializer used to convert between RabbitMQ's
                                       messages and Flink's objects.
        """
        JRMQSource = get_gateway().jvm.org.apache.flink.streaming.connectors.rabbitmq.RMQSource
        j_rmq_source = JRMQSource(
            connection_config._j_rmq_connection_config,
            queue_name,
            use_correlation_id,
            deserialization_schema._j_deserialization_schema
        )
        super(RMQSource, self).__init__(source_func=j_rmq_source)


class RMQSink(SinkFunction):
    def __init__(self, connection_config: 'RMQConnectionConfig',
                 queue_name: str, serialization_schema: SerializationSchema):
        """
        Creates a new RabbitMQ sink.

        :param connection_config: The RabbiMQ connection configuration.
        :param queue_name: The queue to publish messages to.
        :param serialization_schema: A serializer used to convert Flink objects to bytes.
        """
        JRMQSink = get_gateway().jvm.org.apache.flink.streaming.connectors.rabbitmq.RMQSink
        j_rmq_sink = JRMQSink(
            connection_config._j_rmq_connection_config,
            queue_name,
            serialization_schema._j_serialization_schema,
        )
        super(RMQSink, self).__init__(sink_func=j_rmq_sink)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import logging
import sys

from pyflink.common import Types
from pyflink.datastream import StreamExecutionEnvironment
from pyflink.datastream.connectors.kafka import FlinkKafkaProducer, FlinkKafkaConsumer
from pyflink.datastream.formats.avro import AvroRowSerializationSchema, AvroRowDeserializationSchema


# Make sure that the Kafka cluster is started and the topic 'test_avro_topic' is
# created before executing this job.
def write_to_kafka(env):
    ds = env.from_collection([
        (1, 'hi'), (2, 'hello'), (3, 'hi'), (4, 'hello'), (5, 'hi'), (6, 'hello'), (6, 'hello')],
        type_info=Types.ROW([Types.INT(), Types.STRING()]))

    serialization_schema = AvroRowSerializationSchema(
        avro_schema_string="""
            {
                "type": "record",
                "name": "TestRecord",
                "fields": [
                    {"name": "id", "type": "int"},
                    {"name": "name", "type": "string"}
                ]
            }"""
    )

    kafka_producer = FlinkKafkaProducer(
        topic='test_avro_topic',
        serialization_schema=serialization_schema,
        producer_config={'bootstrap.servers': 'localhost:9092', 'group.id': 'test_group'}
    )

    # note that the output type of ds must be RowTypeInfo
    ds.add_sink(kafka_producer)
    env.execute()


def read_from_kafka(env):
    deserialization_schema = AvroRowDeserializationSchema(
        avro_schema_string="""
            {
                "type": "record",
                "name": "TestRecord",
                "fields": [
                    {"name": "id", "type": "int"},
                    {"name": "name", "type": "string"}
                ]
            }"""
    )

    kafka_consumer = FlinkKafkaConsumer(
        topics='test_avro_topic',
        deserialization_schema=deserialization_schema,
        properties={'bootstrap.servers': 'localhost:9092', 'group.id': 'test_group_1'}
    )
    kafka_consumer.set_start_from_earliest()

    env.add_source(kafka_consumer).print()
    env.execute()


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format="%(message)s")

    env = StreamExecutionEnvironment.get_execution_environment()
    env.add_jars("file:///path/to/flink-sql-avro-1.15.0.jar",
                 "file:///path/to/flink-sql-connector-kafka-1.15.0.jar")

    print("start writing data to kafka")
    write_to_kafka(env)

    print("start reading data from kafka")
    read_from_kafka(env)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
# cython: language_level = 3

from apache_beam.coders.coder_impl cimport InputStream as BInputStream
from apache_beam.coders.coder_impl cimport OutputStream as BOutputStream

from pyflink.fn_execution.stream_fast cimport LengthPrefixInputStream, LengthPrefixOutputStream

cdef class BeamInputStream(LengthPrefixInputStream):
    cdef char*_input_data
    cdef size_t _input_buffer_size
    cdef size_t _input_pos
    cdef BInputStream _input_stream
    cdef void _parse_input_stream(self, BInputStream input_stream)

cdef class BeamSizeBasedOutputStream(LengthPrefixOutputStream):
    cdef char*_output_data
    cdef size_t _output_pos
    cdef size_t _output_buffer_size
    cdef BOutputStream _output_stream

    cdef void reset_output_stream(self, BOutputStream output_stream)
    cpdef bint maybe_flush(self)

cdef class BeamTimeBasedOutputStream(BeamSizeBasedOutputStream):
    cdef bint _flush_event
    cdef object _periodic_flusher

    cpdef void notify_flush(self)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from pyflink.java_gateway import get_gateway
from pyflink.util.java_utils import create_url_class_loader

from pyflink.common import Configuration

__all__ = ['EnvironmentSettings']


class EnvironmentSettings(object):
    """
    Defines all parameters that initialize a table environment. Those parameters are used only
    during instantiation of a :class:`~pyflink.table.TableEnvironment` and cannot be changed
    afterwards.

    Example:
    ::

        >>> EnvironmentSettings.new_instance() \\
        ...     .in_streaming_mode() \\
        ...     .with_built_in_catalog_name("my_catalog") \\
        ...     .with_built_in_database_name("my_database") \\
        ...     .build()

    :func:`~EnvironmentSettings.in_streaming_mode` or :func:`~EnvironmentSettings.in_batch_mode`
    might be convenient as shortcuts.
    """

    class Builder(object):
        """
        A builder for :class:`~EnvironmentSettings`.
        """

        def __init__(self):
            gateway = get_gateway()
            self._j_builder = gateway.jvm.EnvironmentSettings.Builder()

        def with_configuration(self, config: Configuration) -> 'EnvironmentSettings.Builder':
            """
            Creates the EnvironmentSetting with specified Configuration.

            :return: EnvironmentSettings.
            """
            self._j_builder = self._j_builder.withConfiguration(config._j_configuration)
            return self

        def in_batch_mode(self) -> 'EnvironmentSettings.Builder':
            """
            Sets that the components should work in a batch mode. Streaming mode by default.

            :return: This object.
            """
            self._j_builder = self._j_builder.inBatchMode()
            return self

        def in_streaming_mode(self) -> 'EnvironmentSettings.Builder':
            """
            Sets that the components should work in a streaming mode. Enabled by default.

            :return: This object.
            """
            self._j_builder = self._j_builder.inStreamingMode()
            return self

        def with_built_in_catalog_name(self, built_in_catalog_name: str) \
                -> 'EnvironmentSettings.Builder':
            """
            Specifies the name of the initial catalog to be created when instantiating
            a :class:`~pyflink.table.TableEnvironment`.

            This catalog is an in-memory catalog that will be used to store all temporary objects
            (e.g. from :func:`~pyflink.table.TableEnvironment.create_temporary_view` or
            :func:`~pyflink.table.TableEnvironment.create_temporary_system_function`) that cannot
            be persisted because they have no serializable representation.

            It will also be the initial value for the current catalog which can be altered via
            :func:`~pyflink.table.TableEnvironment.use_catalog`.

            Default: "default_catalog".

            :param built_in_catalog_name: The specified built-in catalog name.
            :return: This object.
            """
            self._j_builder = self._j_builder.withBuiltInCatalogName(built_in_catalog_name)
            return self

        def with_built_in_database_name(self, built_in_database_name: str) \
                -> 'EnvironmentSettings.Builder':
            """
            Specifies the name of the default database in the initial catalog to be
            created when instantiating a :class:`~pyflink.table.TableEnvironment`.

            This database is an in-memory database that will be used to store all temporary
            objects (e.g. from :func:`~pyflink.table.TableEnvironment.create_temporary_view` or
            :func:`~pyflink.table.TableEnvironment.create_temporary_system_function`) that cannot
            be persisted because they have no serializable representation.

            It will also be the initial value for the current catalog which can be altered via
            :func:`~pyflink.table.TableEnvironment.use_catalog`.

            Default: "default_database".

            :param built_in_database_name: The specified built-in database name.
            :return: This object.
            """
            self._j_builder = self._j_builder.withBuiltInDatabaseName(built_in_database_name)
            return self

        def build(self) -> 'EnvironmentSettings':
            """
            Returns an immutable instance of EnvironmentSettings.

            :return: an immutable instance of EnvironmentSettings.
            """
            gateway = get_gateway()
            context_classloader = gateway.jvm.Thread.currentThread().getContextClassLoader()
            new_classloader = create_url_class_loader([], context_classloader)
            gateway.jvm.Thread.currentThread().setContextClassLoader(new_classloader)
            return EnvironmentSettings(self._j_builder.build())

    def __init__(self, j_environment_settings):
        self._j_environment_settings = j_environment_settings

    def get_built_in_catalog_name(self) -> str:
        """
        Gets the specified name of the initial catalog to be created when instantiating a
        :class:`~pyflink.table.TableEnvironment`.

        :return: The specified name of the initial catalog to be created.
        """
        return self._j_environment_settings.getBuiltInCatalogName()

    def get_built_in_database_name(self) -> str:
        """
        Gets the specified name of the default database in the initial catalog to be created when
        instantiating a :class:`~pyflink.table.TableEnvironment`.

        :return: The specified name of the default database in the initial catalog to be created.
        """
        return self._j_environment_settings.getBuiltInDatabaseName()

    def is_streaming_mode(self) -> bool:
        """
        Tells if the :class:`~pyflink.table.TableEnvironment` should work in a batch or streaming
        mode.

        :return: True if the TableEnvironment should work in a streaming mode, false otherwise.
        """
        return self._j_environment_settings.isStreamingMode()

    def get_configuration(self) -> Configuration:
        """
        Get the underlying `pyflink.common.Configuration`.

        :return: Configuration with specified value.
        """
        return Configuration(j_configuration=self._j_environment_settings.getConfiguration())

    @staticmethod
    def new_instance() -> 'EnvironmentSettings.Builder':
        """
        Creates a builder for creating an instance of EnvironmentSettings.

        :return: A builder of EnvironmentSettings.
        """
        return EnvironmentSettings.Builder()

    @staticmethod
    def in_streaming_mode() -> 'EnvironmentSettings':
        """
        Creates a default instance of EnvironmentSettings in streaming execution mode.

        In this mode, both bounded and unbounded data streams can be processed.

        This method is a shortcut for creating a :class:`~pyflink.table.TableEnvironment` with
        little code. Use the builder provided in :func:`EnvironmentSettings.new_instance` for
        advanced settings.

        :return: EnvironmentSettings.
        """
        return EnvironmentSettings.new_instance().in_streaming_mode().build()

    @staticmethod
    def in_batch_mode() -> 'EnvironmentSettings':
        """
        Creates a default instance of EnvironmentSettings in batch execution mode.

        This mode is highly optimized for batch scenarios. Only bounded data streams can be
        processed in this mode.

        This method is a shortcut for creating a :class:`~pyflink.table.TableEnvironment` with
        little code. Use the builder provided in :func:`EnvironmentSettings.new_instance` for
        advanced settings.

        :return: EnvironmentSettings.
        """
        return EnvironmentSettings.new_instance().in_batch_mode().build()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from typing import Optional

from pyflink.common import Duration
from pyflink.datastream.externalized_checkpoint_retention import ExternalizedCheckpointRetention
from pyflink.datastream.checkpointing_mode import CheckpointingMode

__all__ = ['CheckpointConfig']


class CheckpointConfig(object):
    """
    Configuration that captures all checkpointing related settings.
    """

    def __init__(self, j_checkpoint_config):
        self._j_checkpoint_config = j_checkpoint_config

    def is_checkpointing_enabled(self) -> bool:
        """
        Checks whether checkpointing is enabled.

        :return: True if checkpointing is enables, false otherwise.
        """
        return self._j_checkpoint_config.isCheckpointingEnabled()

    def get_checkpointing_mode(self) -> CheckpointingMode:
        """
        Gets the checkpointing mode (exactly-once vs. at-least-once).

        .. seealso:: :func:`set_checkpointing_mode`

        :return: The :class:`CheckpointingMode`.
        """
        return CheckpointingMode._from_j_checkpointing_mode(
            self._j_checkpoint_config.getCheckpointingConsistencyMode())

    def set_checkpointing_mode(self, checkpointing_mode: CheckpointingMode) -> 'CheckpointConfig':
        """
        Sets the checkpointing mode (:data:`CheckpointingMode.EXACTLY_ONCE` vs.
        :data:`CheckpointingMode.AT_LEAST_ONCE`).

        Example:
        ::

            >>> config.set_checkpointing_mode(CheckpointingMode.AT_LEAST_ONCE)

        :param checkpointing_mode: The :class:`CheckpointingMode`.
        """
        self._j_checkpoint_config.setCheckpointingConsistencyMode(
            CheckpointingMode._to_j_checkpointing_mode(checkpointing_mode))
        return self

    def get_checkpoint_interval(self) -> int:
        """
        Gets the interval in which checkpoints are periodically scheduled.

        This setting defines the base interval. Checkpoint triggering may be delayed by the settings
        :func:`get_max_concurrent_checkpoints` and :func:`get_min_pause_between_checkpoints`.

        :return: The checkpoint interval, in milliseconds.
        """
        return self._j_checkpoint_config.getCheckpointInterval()

    def set_checkpoint_interval(self, checkpoint_interval: int) -> 'CheckpointConfig':
        """
        Sets the interval in which checkpoints are periodically scheduled.

        This setting defines the base interval. Checkpoint triggering may be delayed by the settings
        :func:`set_max_concurrent_checkpoints` and :func:`set_min_pause_between_checkpoints`.

        :param checkpoint_interval: The checkpoint interval, in milliseconds.
        """
        self._j_checkpoint_config.setCheckpointInterval(checkpoint_interval)
        return self

    def get_checkpoint_timeout(self) -> int:
        """
        Gets the maximum time that a checkpoint may take before being discarded.

        :return: The checkpoint timeout, in milliseconds.
        """
        return self._j_checkpoint_config.getCheckpointTimeout()

    def set_checkpoint_timeout(self, checkpoint_timeout: int) -> 'CheckpointConfig':
        """
        Sets the maximum time that a checkpoint may take before being discarded.

        :param checkpoint_timeout: The checkpoint timeout, in milliseconds.
        """
        self._j_checkpoint_config.setCheckpointTimeout(checkpoint_timeout)
        return self

    def get_min_pause_between_checkpoints(self) -> int:
        """
        Gets the minimal pause between checkpointing attempts. This setting defines how soon the
        checkpoint coordinator may trigger another checkpoint after it becomes possible to trigger
        another checkpoint with respect to the maximum number of concurrent checkpoints
        (see :func:`get_max_concurrent_checkpoints`).

        :return: The minimal pause before the next checkpoint is triggered.
        """
        return self._j_checkpoint_config.getMinPauseBetweenCheckpoints()

    def set_min_pause_between_checkpoints(self,
                                          min_pause_between_checkpoints: int) -> 'CheckpointConfig':
        """
        Sets the minimal pause between checkpointing attempts. This setting defines how soon the
        checkpoint coordinator may trigger another checkpoint after it becomes possible to trigger
        another checkpoint with respect to the maximum number of concurrent checkpoints
        (see :func:`set_max_concurrent_checkpoints`).

        If the maximum number of concurrent checkpoints is set to one, this setting makes
        effectively sure that a minimum amount of time passes where no checkpoint is in progress
        at all.

        :param min_pause_between_checkpoints: The minimal pause before the next checkpoint is
                                              triggered.
        """
        self._j_checkpoint_config.setMinPauseBetweenCheckpoints(min_pause_between_checkpoints)
        return self

    def get_max_concurrent_checkpoints(self) -> int:
        """
        Gets the maximum number of checkpoint attempts that may be in progress at the same time.
        If this value is *n*, then no checkpoints will be triggered while *n* checkpoint attempts
        are currently in flight. For the next checkpoint to be triggered, one checkpoint attempt
        would need to finish or expire.

        :return: The maximum number of concurrent checkpoint attempts.
        """
        return self._j_checkpoint_config.getMaxConcurrentCheckpoints()

    def set_max_concurrent_checkpoints(self, max_concurrent_checkpoints: int) -> 'CheckpointConfig':
        """
        Sets the maximum number of checkpoint attempts that may be in progress at the same time.
        If this value is *n*, then no checkpoints will be triggered while *n* checkpoint attempts
        are currently in flight. For the next checkpoint to be triggered, one checkpoint attempt
        would need to finish or expire.

        :param max_concurrent_checkpoints: The maximum number of concurrent checkpoint attempts.
        """
        self._j_checkpoint_config.setMaxConcurrentCheckpoints(max_concurrent_checkpoints)
        return self

    def is_fail_on_checkpointing_errors(self) -> bool:
        """
        This determines the behaviour of tasks if there is an error in their local checkpointing.
        If this returns true, tasks will fail as a reaction. If this returns false, task will only
        decline the failed checkpoint.

        :return: ``True`` if failing on checkpointing errors, false otherwise.

        .. note:: Deprecated in 2.0. Use :func:`get_tolerable_checkpoint_failure_number` instead.
        """
        return self.get_tolerable_checkpoint_failure_number() == 0

    def set_fail_on_checkpointing_errors(self,
                                         fail_on_checkpointing_errors: bool) -> 'CheckpointConfig':
        """
        Sets the expected behaviour for tasks in case that they encounter an error in their
        checkpointing procedure. If this is set to true, the task will fail on checkpointing error.
        If this is set to false, the task will only decline a the checkpoint and continue running.
        The default is true.

        Example:
        ::

            >>> config.set_fail_on_checkpointing_errors(False)

        :param fail_on_checkpointing_errors: ``True`` if failing on checkpointing errors,
                                             false otherwise.

        .. note:: Deprecated in 2.0. Use :func:`set_tolerable_checkpoint_failure_number` instead.
        """
        if fail_on_checkpointing_errors:
            self.set_tolerable_checkpoint_failure_number(0)
        else:
            self.set_tolerable_checkpoint_failure_number(2147483647)
        return self

    def get_tolerable_checkpoint_failure_number(self) -> int:
        """
        Get the defined number of consecutive checkpoint failures that will be tolerated, before the
        whole job is failed over.

        :return: The maximum number of tolerated checkpoint failures.
        """
        return self._j_checkpoint_config.getTolerableCheckpointFailureNumber()

    def set_tolerable_checkpoint_failure_number(self,
                                                tolerable_checkpoint_failure_number: int
                                                ) -> 'CheckpointConfig':
        """
        This defines how many consecutive checkpoint failures will be tolerated, before the whole
        job is failed over. The default value is `0`, which means no checkpoint failures will be
        tolerated, and the job will fail on first reported checkpoint failure.

        Example:
        ::

            >>> config.set_tolerable_checkpoint_failure_number(2)

        :param tolerable_checkpoint_failure_number: The maximum number of tolerated checkpoint
                                                    failures.
        """
        self._j_checkpoint_config.setTolerableCheckpointFailureNumber(
            tolerable_checkpoint_failure_number)
        return self

    def set_externalized_checkpoint_retention(
            self,
            retention_mode: 'ExternalizedCheckpointRetention') -> 'CheckpointConfig':
        """
        Sets the mode for externalized checkpoint clean-up. Externalized checkpoints will be enabled
        automatically unless the mode is set to
        :data:`ExternalizedCheckpointRetention.NO_EXTERNALIZED_CHECKPOINTS`.

        Externalized checkpoints write their meta data out to persistent storage and are **not**
        automatically cleaned up when the owning job fails or is suspended (terminating with job
        status ``FAILED`` or ``SUSPENDED``). In this case, you have to manually clean up the
        checkpoint state, both the meta data and actual program state.

        The :class:`ExternalizedCheckpointRetention` mode defines how an externalized checkpoint
        should be cleaned up on job cancellation. If you choose to retain externalized checkpoints
        on cancellation you have to handle checkpoint clean-up manually when you cancel the job as
        well (terminating with job status ``CANCELED``).

        The target directory for externalized checkpoints is configured via
        ``org.apache.flink.configuration.CheckpointingOptions#CHECKPOINTS_DIRECTORY``.

        Example:
        ::

            >>> config.set_externalized_checkpoint_retention(
            ...     ExternalizedCheckpointRetention.RETAIN_ON_CANCELLATION)

        :param retention_mode: Externalized checkpoint clean-up behaviour, the mode could be
                             :data:`ExternalizedCheckpointRetention.DELETE_ON_CANCELLATION`,
                             :data:`ExternalizedCheckpointRetention.RETAIN_ON_CANCELLATION` or
                             :data:`ExternalizedCheckpointRetention.NO_EXTERNALIZED_CHECKPOINTS`
        """
        self._j_checkpoint_config.setExternalizedCheckpointRetention(
            ExternalizedCheckpointRetention._to_j_externalized_checkpoint_retention(retention_mode))
        return self

    def is_externalized_checkpoints_enabled(self) -> bool:
        """
        Returns whether checkpoints should be persisted externally.

        :return: ``True`` if checkpoints should be externalized, false otherwise.
        """
        return self._j_checkpoint_config.isExternalizedCheckpointsEnabled()

    def get_externalized_checkpoint_retention(self) -> Optional['ExternalizedCheckpointRetention']:
        """
        Returns the cleanup behaviour for externalized checkpoints.

        :return: The cleanup behaviour for externalized checkpoints or ``None`` if none is
                 configured.
        """
        retention_mode = self._j_checkpoint_config.getExternalizedCheckpointRetention()
        if retention_mode is None:
            return None
        else:
            return ExternalizedCheckpointRetention._from_j_externalized_checkpoint_retention(
                retention_mode)

    def is_unaligned_checkpoints_enabled(self) -> bool:
        """
        Returns whether unaligned checkpoints are enabled.

        :return: ``True`` if unaligned checkpoints are enabled.
        """
        return self._j_checkpoint_config.isUnalignedCheckpointsEnabled()

    def enable_unaligned_checkpoints(self, enabled: bool = True) -> 'CheckpointConfig':
        """
        Enables unaligned checkpoints, which greatly reduce checkpointing times under backpressure.

        Unaligned checkpoints contain data stored in buffers as part of the checkpoint state, which
        allows checkpoint barriers to overtake these buffers. Thus, the checkpoint duration becomes
        independent of the current throughput as checkpoint barriers are effectively not embedded
        into the stream of data anymore.

        Unaligned checkpoints can only be enabled if :func:`get_checkpointing_mode` is
        :data:`CheckpointingMode.EXACTLY_ONCE`.

        :param enabled: ``True`` if a checkpoints should be taken in unaligned mode.
        """
        self._j_checkpoint_config.enableUnalignedCheckpoints(enabled)
        return self

    def disable_unaligned_checkpoints(self) -> 'CheckpointConfig':
        """
        Enables unaligned checkpoints, which greatly reduce checkpointing times under backpressure
        (experimental).

        Unaligned checkpoints contain data stored in buffers as part of the checkpoint state, which
        allows checkpoint barriers to overtake these buffers. Thus, the checkpoint duration becomes
        independent of the current throughput as checkpoint barriers are effectively not embedded
        into the stream of data anymore.

        Unaligned checkpoints can only be enabled if :func:`get_checkpointing_mode` is
        :data:`CheckpointingMode.EXACTLY_ONCE`.
        """
        self.enable_unaligned_checkpoints(False)
        return self

    def set_aligned_checkpoint_timeout(self, alignment_timeout: Duration) -> 'CheckpointConfig':
        """
        Only relevant if :func:`enable_unaligned_checkpoints` is enabled.

        If ``alignment_timeout`` has value equal to ``0``, checkpoints will always start unaligned.
        If ``alignment_timeout`` has value greater then ``0``, checkpoints will start aligned. If
        during checkpointing, checkpoint start delay exceeds this ``alignment_timeout``, alignment
        will timeout and checkpoint will start working as unaligned checkpoint.

        :param alignment_timeout: The duration until the aligned checkpoint will be converted into
                                  an unaligned checkpoint.
        """
        self._j_checkpoint_config.setAlignedCheckpointTimeout(alignment_timeout._j_duration)
        return self

    def get_aligned_checkpoint_timeout(self) -> 'Duration':
        """
        Returns the alignment timeout, as configured via :func:`set_alignment_timeout` or
        ``org.apache.flink.configuration.CheckpointingOptions#ALIGNED_CHECKPOINT_TIMEOUT``.

        :return: the alignment timeout.
        """
        return Duration(self._j_checkpoint_config.getAlignedCheckpointTimeout())

    def set_alignment_timeout(self, alignment_timeout: Duration) -> 'CheckpointConfig':
        """
        Only relevant if :func:`enable_unaligned_checkpoints` is enabled.

        If ``alignment_timeout`` has value equal to ``0``, checkpoints will always start unaligned.
        If ``alignment_timeout`` has value greater then ``0``, checkpoints will start aligned. If
        during checkpointing, checkpoint start delay exceeds this ``alignment_timeout``, alignment
        will timeout and checkpoint will start working as unaligned checkpoint.

        :param alignment_timeout: The duration until the aligned checkpoint will be converted into
                                  an unaligned checkpoint.

        .. note:: Deprecated in 2.0. Use :func:`set_aligned_checkpoint_timeout` instead.
        """
        self.set_aligned_checkpoint_timeout(alignment_timeout)
        return self

    def get_alignment_timeout(self) -> 'Duration':
        """
        Returns the alignment timeout, as configured via :func:`set_alignment_timeout` or
        ``org.apache.flink.configuration.CheckpointingOptions#ALIGNED_CHECKPOINT_TIMEOUT``.

        :return: the alignment timeout.

        .. note:: Deprecated in 2.0. Use :func:`get_aligned_checkpoint_timeout` instead.
        """
        return self.get_aligned_checkpoint_timeout()

    def set_force_unaligned_checkpoints(
            self,
            force_unaligned_checkpoints: bool = True) -> 'CheckpointConfig':
        """
        Checks whether unaligned checkpoints are forced, despite currently non-checkpointable
        iteration feedback or custom partitioners.

        :param force_unaligned_checkpoints: The flag to force unaligned checkpoints.
        """
        self._j_checkpoint_config.setForceUnalignedCheckpoints(force_unaligned_checkpoints)
        return self

    def is_force_unaligned_checkpoints(self) -> 'bool':
        """
        Checks whether unaligned checkpoints are forced, despite iteration feedback or custom
        partitioners.

        :return: True, if unaligned checkpoints are forced, false otherwise.
        """
        return self._j_checkpoint_config.isForceUnalignedCheckpoints()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from pyflink.common.utils import JavaObjectWrapper
from pyflink.java_gateway import get_gateway

__all__ = [
    'SerializationSchema',
    'DeserializationSchema',
    'SimpleStringSchema',
    'ByteArraySchema',
    'Encoder',
    'BulkWriterFactory'
]


class SerializationSchema(object):
    """
    Base class for SerializationSchema. The serialization schema describes how to turn a data object
    into a different serialized representation. Most data sinks (for example Apache Kafka) require
    the data to be handed to them in a specific format (for example as byte strings).
    """
    def __init__(self, j_serialization_schema=None):
        self._j_serialization_schema = j_serialization_schema


class DeserializationSchema(object):
    """
    Base class for DeserializationSchema. The deserialization schema describes how to turn the byte
    messages delivered by certain data sources (for example Apache Kafka) into data types (Java/
    Scala objects) that are processed by Flink.

    In addition, the DeserializationSchema describes the produced type which lets Flink create
    internal serializers and structures to handle the type.
    """
    def __init__(self, j_deserialization_schema=None):
        self._j_deserialization_schema = j_deserialization_schema


class SimpleStringSchema(SerializationSchema, DeserializationSchema):
    """
    Very simple serialization/deserialization schema for strings. By default, the serializer uses
    'UTF-8' for string/byte conversion.
    """

    def __init__(self, charset: str = 'UTF-8'):
        gate_way = get_gateway()
        j_char_set = gate_way.jvm.java.nio.charset.Charset.forName(charset)
        j_simple_string_serialization_schema = gate_way \
            .jvm.org.apache.flink.api.common.serialization.SimpleStringSchema(j_char_set)
        SerializationSchema.__init__(self,
                                     j_serialization_schema=j_simple_string_serialization_schema)
        DeserializationSchema.__init__(
            self, j_deserialization_schema=j_simple_string_serialization_schema)


class ByteArraySchema(SerializationSchema, DeserializationSchema):
    """
    Simple serialization/deserialization schema for bytes.
    """

    def __init__(self):
        gate_way = get_gateway()
        j_simple_byte_serialization_schema = gate_way \
            .jvm.org.apache.flink.api.common.serialization.ByteArraySchema()
        SerializationSchema.__init__(self,
                                     j_serialization_schema=j_simple_byte_serialization_schema)
        DeserializationSchema.__init__(
            self, j_deserialization_schema=j_simple_byte_serialization_schema)


class Encoder(object):
    """
    Encoder is used by the file sink to perform the actual writing of the
    incoming elements to the files in a bucket.
    """

    def __init__(self, j_encoder):
        self._j_encoder = j_encoder

    @staticmethod
    def simple_string_encoder(charset_name: str = "UTF-8") -> 'Encoder':
        """
        A simple Encoder that uses toString() on the input elements and writes them to
        the output bucket file separated by newline.
        """
        j_encoder = get_gateway().jvm.org.apache.flink.api.common.serialization.\
            SimpleStringEncoder(charset_name)
        return Encoder(j_encoder)


class BulkWriterFactory(JavaObjectWrapper):
    """
    The Python wrapper of Java BulkWriter.Factory interface, which is the base interface for data
    sinks that write records into files in a bulk manner.
    """

    def __init__(self, j_bulk_writer_factory):
        super().__init__(j_bulk_writer_factory)


class RowDataBulkWriterFactory(BulkWriterFactory):
    """
    A :class:`～BulkWriterFactory` that receives records with RowData type. This is for indicating
    that Row record from Python must be first converted to RowData.
    """

    def __init__(self, j_bulk_writer_factory, row_type):
        super().__init__(j_bulk_writer_factory)
        self._row_type = row_type

    def get_row_type(self):
        return self._row_type



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Iterable, List, Any, Iterator, Dict, Tuple

T = TypeVar('T')
K = TypeVar('K')
V = TypeVar('V')

__all__ = ['DataView', 'ListView', 'MapView']


class DataView(ABC):
    """
    A DataView is a collection type that can be used in the accumulator of an user defined
    :class:`pyflink.table.AggregateFunction`. Depending on the context in which the function
    is used, a DataView can be backed by a normal collection or a state backend.
    """

    @abstractmethod
    def clear(self) -> None:
        """
        Clears the DataView and removes all data.
        """
        pass


class ListView(DataView, Generic[T]):
    """
    A :class:`DataView` that provides list-like functionality in the accumulator of an
    AggregateFunction when large amounts of data are expected.
    """

    def __init__(self):
        self._list = []

    def get(self) -> Iterable[T]:
        """
        Returns an iterable of this list view.
        """
        return self._list

    def add(self, value: T) -> None:
        """
        Adds the given value to this list view.
        """
        self._list.append(value)

    def add_all(self, values: List[T]) -> None:
        """
        Adds all of the elements of the specified list to this list view.
        """
        self._list.extend(values)

    def clear(self) -> None:
        self._list = []

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ListView):
            iter_obj = other.get()
            self_iterator = iter(self)
            for value in iter_obj:
                try:
                    self_value = next(self_iterator)
                except StopIteration:
                    # this list view is shorter than another one
                    return False
                if self_value != value:
                    # the elements are not the same.
                    return False
            try:
                next(self_iterator)
            except StopIteration:
                # the length of this list view is the same as another one
                return True
            else:
                # this list view is longer than another one
                return False
        else:
            # the object is not a ListView
            return False

    def __hash__(self) -> int:
        return hash(self._list)

    def __iter__(self) -> Iterator[T]:
        return iter(self.get())


class MapView(Generic[K, V]):
    """
    A :class:`DataView` that provides dict-like functionality in the accumulator of an
    AggregateFunction when large amounts of data are expected.
    """

    def __init__(self):
        self._dict = dict()

    def get(self, key: K) -> V:
        """
        Return the value for the specified key.
        """
        return self._dict[key]

    def put(self, key: K, value: V) -> None:
        """
        Inserts a value for the given key into the map view.
        If the map view already contains a value for the key, the existing value is overwritten.
        """
        self._dict[key] = value

    def put_all(self, dict_value: Dict[K, V]) -> None:
        """
        Inserts all mappings from the specified map to this map view.
        """
        self._dict.update(dict_value)

    def remove(self, key: K) -> None:
        """
        Deletes the value for the given key.
        """
        del self._dict[key]

    def contains(self, key: K) -> bool:
        """
        Checks if the map view contains a value for a given key.
        """
        return key in self._dict

    def items(self) -> Iterable[Tuple[K, V]]:
        """
        Returns all entries of the map view.
        """
        return self._dict.items()

    def keys(self) -> Iterable[K]:
        """
        Returns all the keys in the map view.
        """
        return self._dict.keys()

    def values(self) -> Iterable[V]:
        """
        Returns all the values in the map view.
        """
        return self._dict.values()

    def is_empty(self) -> bool:
        """
        Returns true if the map view contains no key-value mappings, otherwise false.
        """
        return len(self._dict) == 0

    def clear(self) -> None:
        """
        Removes all entries of this map.
        """
        self._dict.clear()

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        if other.__class__ == MapView:
            return self._dict == other._dict
        else:
            # comparing the content of state backed map view is too expensive
            return other is self

    def __getitem__(self, key: K) -> V:
        return self.get(key)

    def __setitem__(self, key: K, value: V) -> None:
        self.put(key, value)

    def __delitem__(self, key: K) -> None:
        self.remove(key)

    def __contains__(self, key: K) -> bool:
        return self.contains(key)

    def __iter__(self) -> Iterator[K]:
        return iter(self.keys())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC

from pyflink.datastream import (ProcessFunction, KeyedProcessFunction, CoProcessFunction,
                                KeyedCoProcessFunction, TimerService, TimeDomain)
from pyflink.datastream.functions import (BaseBroadcastProcessFunction, BroadcastProcessFunction,
                                          KeyedBroadcastProcessFunction)
from pyflink.datastream.state import MapStateDescriptor, BroadcastState, ReadOnlyBroadcastState
from pyflink.fn_execution.datastream.embedded.state_impl import (ReadOnlyBroadcastStateImpl,
                                                                 BroadcastStateImpl)
from pyflink.fn_execution.datastream.embedded.timerservice_impl import TimerServiceImpl
from pyflink.fn_execution.embedded.converters import from_type_info_proto, from_type_info
from pyflink.fn_execution.embedded.java_utils import to_java_state_descriptor


class InternalProcessFunctionContext(ProcessFunction.Context, CoProcessFunction.Context,
                                     TimerService):
    def __init__(self, j_context):
        self._j_context = j_context

    def timer_service(self) -> TimerService:
        return self

    def timestamp(self) -> int:
        return self._j_context.timestamp()

    def current_processing_time(self):
        return self._j_context.currentProcessingTime()

    def current_watermark(self):
        return self._j_context.currentWatermark()

    def register_processing_time_timer(self, timestamp: int):
        raise Exception("Register timers is only supported on a keyed stream.")

    def register_event_time_timer(self, timestamp: int):
        raise Exception("Register timers is only supported on a keyed stream.")

    def delete_processing_time_timer(self, t: int):
        raise Exception("Deleting timers is only supported on a keyed streams.")

    def delete_event_time_timer(self, t: int):
        raise Exception("Deleting timers is only supported on a keyed streams.")


class InternalKeyedProcessFunctionContext(KeyedProcessFunction.Context,
                                          KeyedCoProcessFunction.Context):

    def __init__(self, j_context, key_type_info):
        self._j_context = j_context
        self._timer_service = TimerServiceImpl(self._j_context.timerService())
        self._key_converter = from_type_info_proto(key_type_info)

    def get_current_key(self):
        return self._key_converter.to_internal(self._j_context.getCurrentKey())

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._j_context.timestamp()


class InternalKeyedProcessFunctionOnTimerContext(KeyedProcessFunction.OnTimerContext,
                                                 KeyedProcessFunction.Context,
                                                 KeyedCoProcessFunction.OnTimerContext,
                                                 KeyedCoProcessFunction.Context):

    def __init__(self, j_timer_context, key_type_info):
        self._j_timer_context = j_timer_context
        self._timer_service = TimerServiceImpl(self._j_timer_context.timerService())
        self._key_converter = from_type_info_proto(key_type_info)

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._j_timer_context.timestamp()

    def time_domain(self) -> TimeDomain:
        return TimeDomain(self._j_timer_context.timeDomain())

    def get_current_key(self):
        return self._key_converter.to_internal(self._j_timer_context.getCurrentKey())


class InternalWindowTimerContext(object):
    def __init__(self, j_timer_context, key_type_info, window_converter):
        self._j_timer_context = j_timer_context
        self._key_converter = from_type_info_proto(key_type_info)
        self._window_converter = window_converter

    def timestamp(self) -> int:
        return self._j_timer_context.timestamp()

    def window(self):
        return self._window_converter.to_internal(self._j_timer_context.getWindow())

    def get_current_key(self):
        return self._key_converter.to_internal(self._j_timer_context.getCurrentKey())


class InternalBaseBroadcastProcessFunctionContext(BaseBroadcastProcessFunction.Context, ABC):

    def __init__(self, j_context, j_operator_state_backend):
        self._j_context = j_context
        self._j_operator_state_backend = j_operator_state_backend

    def timestamp(self) -> int:
        return self._j_context.timestamp()

    def current_processing_time(self) -> int:
        return self._j_context.currentProcessingTime()

    def current_watermark(self) -> int:
        return self._j_context.currentWatermark()


class InternalBroadcastProcessFunctionContext(InternalBaseBroadcastProcessFunctionContext,
                                              BroadcastProcessFunction.Context):

    def __init__(self, j_context, j_operator_state_backend):
        super(InternalBroadcastProcessFunctionContext, self).__init__(
            j_context, j_operator_state_backend)

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> BroadcastState:
        return BroadcastStateImpl(
            self._j_operator_state_backend.getBroadcastState(
                to_java_state_descriptor(state_descriptor)),
            from_type_info(state_descriptor.type_info))


class InternalBroadcastProcessFunctionReadOnlyContext(InternalBaseBroadcastProcessFunctionContext,
                                                      BroadcastProcessFunction.ReadOnlyContext):

    def __init__(self, j_context, j_operator_state_backend):
        super(InternalBroadcastProcessFunctionReadOnlyContext, self).__init__(
            j_context, j_operator_state_backend)

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return ReadOnlyBroadcastStateImpl(
            self._j_operator_state_backend.getBroadcastState(
                to_java_state_descriptor(state_descriptor)),
            from_type_info(state_descriptor.type_info))


class InternalKeyedBroadcastProcessFunctionContext(InternalBaseBroadcastProcessFunctionContext,
                                                   KeyedBroadcastProcessFunction.Context):

    def __init__(self, j_context, j_operator_state_backend):
        super(InternalKeyedBroadcastProcessFunctionContext, self).__init__(
            j_context, j_operator_state_backend)

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> BroadcastState:
        return BroadcastStateImpl(
            self._j_operator_state_backend.getBroadcastState(
                to_java_state_descriptor(state_descriptor)),
            from_type_info(state_descriptor.type_info))


class InternalKeyedBroadcastProcessFunctionReadOnlyContext(
    InternalBaseBroadcastProcessFunctionContext,
    KeyedBroadcastProcessFunction.ReadOnlyContext
):

    def __init__(self, j_context, key_type_info, j_operator_state_backend):
        super(InternalKeyedBroadcastProcessFunctionReadOnlyContext, self).__init__(
            j_context, j_operator_state_backend)
        self._key_converter = from_type_info_proto(key_type_info)
        self._timer_service = TimerServiceImpl(self._j_context.timerService())

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return ReadOnlyBroadcastStateImpl(
            self._j_operator_state_backend.getBroadcastState(
                to_java_state_descriptor(state_descriptor)),
            from_type_info(state_descriptor.type_info))

    def timer_service(self) -> TimerService:
        return self._timer_service

    def get_current_key(self):
        return self._key_converter.to_internal(self._j_context.getCurrentKey())


class InternalKeyedBroadcastProcessFunctionOnTimerContext(
    InternalBaseBroadcastProcessFunctionContext,
    KeyedBroadcastProcessFunction.OnTimerContext,
):

    def __init__(self, j_timer_context, key_type_info, j_operator_state_backend):
        super(InternalKeyedBroadcastProcessFunctionOnTimerContext, self).__init__(
            j_timer_context, j_operator_state_backend)
        self._timer_service = TimerServiceImpl(self._j_context.timerService())
        self._key_converter = from_type_info_proto(key_type_info)

    def get_broadcast_state(self, state_descriptor: MapStateDescriptor) -> ReadOnlyBroadcastState:
        return ReadOnlyBroadcastStateImpl(
            self._j_operator_state_backend.getBroadcastState(
                to_java_state_descriptor(state_descriptor)),
            from_type_info(state_descriptor.type_info))

    def current_processing_time(self) -> int:
        return self._timer_service.current_processing_time()

    def current_watermark(self) -> int:
        return self._timer_service.current_watermark()

    def timer_service(self) -> TimerService:
        return self._timer_service

    def timestamp(self) -> int:
        return self._j_context.timestamp()

    def time_domain(self) -> TimeDomain:
        return TimeDomain(self._j_context.timeDomain())

    def get_current_key(self):
        return self._key_converter.to_internal(self._j_context.getCurrentKey())



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CoderInfoDescriptor(_message.Message):
    __slots__ = ["arrow_type", "flatten_row_type", "mode", "over_window_arrow_type", "raw_type", "row_type", "separated_with_end_message"]
    class Mode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class ArrowType(_message.Message):
        __slots__ = ["schema"]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        schema: Schema
        def __init__(self, schema: _Optional[_Union[Schema, _Mapping]] = ...) -> None: ...
    class FlattenRowType(_message.Message):
        __slots__ = ["schema"]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        schema: Schema
        def __init__(self, schema: _Optional[_Union[Schema, _Mapping]] = ...) -> None: ...
    class OverWindowArrowType(_message.Message):
        __slots__ = ["schema"]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        schema: Schema
        def __init__(self, schema: _Optional[_Union[Schema, _Mapping]] = ...) -> None: ...
    class RawType(_message.Message):
        __slots__ = ["type_info"]
        TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
        type_info: TypeInfo
        def __init__(self, type_info: _Optional[_Union[TypeInfo, _Mapping]] = ...) -> None: ...
    class RowType(_message.Message):
        __slots__ = ["schema"]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        schema: Schema
        def __init__(self, schema: _Optional[_Union[Schema, _Mapping]] = ...) -> None: ...
    ARROW_TYPE_FIELD_NUMBER: _ClassVar[int]
    FLATTEN_ROW_TYPE_FIELD_NUMBER: _ClassVar[int]
    MODE_FIELD_NUMBER: _ClassVar[int]
    MULTIPLE: CoderInfoDescriptor.Mode
    OVER_WINDOW_ARROW_TYPE_FIELD_NUMBER: _ClassVar[int]
    RAW_TYPE_FIELD_NUMBER: _ClassVar[int]
    ROW_TYPE_FIELD_NUMBER: _ClassVar[int]
    SEPARATED_WITH_END_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    SINGLE: CoderInfoDescriptor.Mode
    arrow_type: CoderInfoDescriptor.ArrowType
    flatten_row_type: CoderInfoDescriptor.FlattenRowType
    mode: CoderInfoDescriptor.Mode
    over_window_arrow_type: CoderInfoDescriptor.OverWindowArrowType
    raw_type: CoderInfoDescriptor.RawType
    row_type: CoderInfoDescriptor.RowType
    separated_with_end_message: bool
    def __init__(self, flatten_row_type: _Optional[_Union[CoderInfoDescriptor.FlattenRowType, _Mapping]] = ..., row_type: _Optional[_Union[CoderInfoDescriptor.RowType, _Mapping]] = ..., arrow_type: _Optional[_Union[CoderInfoDescriptor.ArrowType, _Mapping]] = ..., over_window_arrow_type: _Optional[_Union[CoderInfoDescriptor.OverWindowArrowType, _Mapping]] = ..., raw_type: _Optional[_Union[CoderInfoDescriptor.RawType, _Mapping]] = ..., mode: _Optional[_Union[CoderInfoDescriptor.Mode, str]] = ..., separated_with_end_message: bool = ...) -> None: ...

class GroupWindow(_message.Message):
    __slots__ = ["allowedLateness", "is_row_time", "is_time_window", "namedProperties", "shift_timezone", "time_field_index", "window_gap", "window_size", "window_slide", "window_type"]
    class WindowProperty(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class WindowType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    ALLOWEDLATENESS_FIELD_NUMBER: _ClassVar[int]
    IS_ROW_TIME_FIELD_NUMBER: _ClassVar[int]
    IS_TIME_WINDOW_FIELD_NUMBER: _ClassVar[int]
    NAMEDPROPERTIES_FIELD_NUMBER: _ClassVar[int]
    PROC_TIME_ATTRIBUTE: GroupWindow.WindowProperty
    ROW_TIME_ATTRIBUTE: GroupWindow.WindowProperty
    SESSION_GROUP_WINDOW: GroupWindow.WindowType
    SHIFT_TIMEZONE_FIELD_NUMBER: _ClassVar[int]
    SLIDING_GROUP_WINDOW: GroupWindow.WindowType
    TIME_FIELD_INDEX_FIELD_NUMBER: _ClassVar[int]
    TUMBLING_GROUP_WINDOW: GroupWindow.WindowType
    WINDOW_END: GroupWindow.WindowProperty
    WINDOW_GAP_FIELD_NUMBER: _ClassVar[int]
    WINDOW_SIZE_FIELD_NUMBER: _ClassVar[int]
    WINDOW_SLIDE_FIELD_NUMBER: _ClassVar[int]
    WINDOW_START: GroupWindow.WindowProperty
    WINDOW_TYPE_FIELD_NUMBER: _ClassVar[int]
    allowedLateness: int
    is_row_time: bool
    is_time_window: bool
    namedProperties: _containers.RepeatedScalarFieldContainer[GroupWindow.WindowProperty]
    shift_timezone: str
    time_field_index: int
    window_gap: int
    window_size: int
    window_slide: int
    window_type: GroupWindow.WindowType
    def __init__(self, window_type: _Optional[_Union[GroupWindow.WindowType, str]] = ..., is_time_window: bool = ..., window_slide: _Optional[int] = ..., window_size: _Optional[int] = ..., window_gap: _Optional[int] = ..., is_row_time: bool = ..., time_field_index: _Optional[int] = ..., allowedLateness: _Optional[int] = ..., namedProperties: _Optional[_Iterable[_Union[GroupWindow.WindowProperty, str]]] = ..., shift_timezone: _Optional[str] = ...) -> None: ...

class Input(_message.Message):
    __slots__ = ["inputConstant", "inputOffset", "udf"]
    INPUTCONSTANT_FIELD_NUMBER: _ClassVar[int]
    INPUTOFFSET_FIELD_NUMBER: _ClassVar[int]
    UDF_FIELD_NUMBER: _ClassVar[int]
    inputConstant: bytes
    inputOffset: int
    udf: UserDefinedFunction
    def __init__(self, udf: _Optional[_Union[UserDefinedFunction, _Mapping]] = ..., inputOffset: _Optional[int] = ..., inputConstant: _Optional[bytes] = ...) -> None: ...

class JobParameter(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class OverWindow(_message.Message):
    __slots__ = ["lower_boundary", "upper_boundary", "window_type"]
    class WindowType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    LOWER_BOUNDARY_FIELD_NUMBER: _ClassVar[int]
    RANGE_SLIDING: OverWindow.WindowType
    RANGE_UNBOUNDED: OverWindow.WindowType
    RANGE_UNBOUNDED_FOLLOWING: OverWindow.WindowType
    RANGE_UNBOUNDED_PRECEDING: OverWindow.WindowType
    ROW_SLIDING: OverWindow.WindowType
    ROW_UNBOUNDED: OverWindow.WindowType
    ROW_UNBOUNDED_FOLLOWING: OverWindow.WindowType
    ROW_UNBOUNDED_PRECEDING: OverWindow.WindowType
    UPPER_BOUNDARY_FIELD_NUMBER: _ClassVar[int]
    WINDOW_TYPE_FIELD_NUMBER: _ClassVar[int]
    lower_boundary: int
    upper_boundary: int
    window_type: OverWindow.WindowType
    def __init__(self, window_type: _Optional[_Union[OverWindow.WindowType, str]] = ..., lower_boundary: _Optional[int] = ..., upper_boundary: _Optional[int] = ...) -> None: ...

class Schema(_message.Message):
    __slots__ = ["fields"]
    class TypeName(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class BinaryInfo(_message.Message):
        __slots__ = ["length"]
        LENGTH_FIELD_NUMBER: _ClassVar[int]
        length: int
        def __init__(self, length: _Optional[int] = ...) -> None: ...
    class CharInfo(_message.Message):
        __slots__ = ["length"]
        LENGTH_FIELD_NUMBER: _ClassVar[int]
        length: int
        def __init__(self, length: _Optional[int] = ...) -> None: ...
    class DecimalInfo(_message.Message):
        __slots__ = ["precision", "scale"]
        PRECISION_FIELD_NUMBER: _ClassVar[int]
        SCALE_FIELD_NUMBER: _ClassVar[int]
        precision: int
        scale: int
        def __init__(self, precision: _Optional[int] = ..., scale: _Optional[int] = ...) -> None: ...
    class Field(_message.Message):
        __slots__ = ["description", "name", "type"]
        DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        TYPE_FIELD_NUMBER: _ClassVar[int]
        description: str
        name: str
        type: Schema.FieldType
        def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., type: _Optional[_Union[Schema.FieldType, _Mapping]] = ...) -> None: ...
    class FieldType(_message.Message):
        __slots__ = ["binary_info", "char_info", "collection_element_type", "decimal_info", "local_zoned_timestamp_info", "map_info", "nullable", "row_schema", "time_info", "timestamp_info", "type_name", "var_binary_info", "var_char_info", "zoned_timestamp_info"]
        BINARY_INFO_FIELD_NUMBER: _ClassVar[int]
        CHAR_INFO_FIELD_NUMBER: _ClassVar[int]
        COLLECTION_ELEMENT_TYPE_FIELD_NUMBER: _ClassVar[int]
        DECIMAL_INFO_FIELD_NUMBER: _ClassVar[int]
        LOCAL_ZONED_TIMESTAMP_INFO_FIELD_NUMBER: _ClassVar[int]
        MAP_INFO_FIELD_NUMBER: _ClassVar[int]
        NULLABLE_FIELD_NUMBER: _ClassVar[int]
        ROW_SCHEMA_FIELD_NUMBER: _ClassVar[int]
        TIMESTAMP_INFO_FIELD_NUMBER: _ClassVar[int]
        TIME_INFO_FIELD_NUMBER: _ClassVar[int]
        TYPE_NAME_FIELD_NUMBER: _ClassVar[int]
        VAR_BINARY_INFO_FIELD_NUMBER: _ClassVar[int]
        VAR_CHAR_INFO_FIELD_NUMBER: _ClassVar[int]
        ZONED_TIMESTAMP_INFO_FIELD_NUMBER: _ClassVar[int]
        binary_info: Schema.BinaryInfo
        char_info: Schema.CharInfo
        collection_element_type: Schema.FieldType
        decimal_info: Schema.DecimalInfo
        local_zoned_timestamp_info: Schema.LocalZonedTimestampInfo
        map_info: Schema.MapInfo
        nullable: bool
        row_schema: Schema
        time_info: Schema.TimeInfo
        timestamp_info: Schema.TimestampInfo
        type_name: Schema.TypeName
        var_binary_info: Schema.VarBinaryInfo
        var_char_info: Schema.VarCharInfo
        zoned_timestamp_info: Schema.ZonedTimestampInfo
        def __init__(self, type_name: _Optional[_Union[Schema.TypeName, str]] = ..., nullable: bool = ..., collection_element_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ..., map_info: _Optional[_Union[Schema.MapInfo, _Mapping]] = ..., row_schema: _Optional[_Union[Schema, _Mapping]] = ..., decimal_info: _Optional[_Union[Schema.DecimalInfo, _Mapping]] = ..., time_info: _Optional[_Union[Schema.TimeInfo, _Mapping]] = ..., timestamp_info: _Optional[_Union[Schema.TimestampInfo, _Mapping]] = ..., local_zoned_timestamp_info: _Optional[_Union[Schema.LocalZonedTimestampInfo, _Mapping]] = ..., zoned_timestamp_info: _Optional[_Union[Schema.ZonedTimestampInfo, _Mapping]] = ..., binary_info: _Optional[_Union[Schema.BinaryInfo, _Mapping]] = ..., var_binary_info: _Optional[_Union[Schema.VarBinaryInfo, _Mapping]] = ..., char_info: _Optional[_Union[Schema.CharInfo, _Mapping]] = ..., var_char_info: _Optional[_Union[Schema.VarCharInfo, _Mapping]] = ...) -> None: ...
    class LocalZonedTimestampInfo(_message.Message):
        __slots__ = ["precision"]
        PRECISION_FIELD_NUMBER: _ClassVar[int]
        precision: int
        def __init__(self, precision: _Optional[int] = ...) -> None: ...
    class MapInfo(_message.Message):
        __slots__ = ["key_type", "value_type"]
        KEY_TYPE_FIELD_NUMBER: _ClassVar[int]
        VALUE_TYPE_FIELD_NUMBER: _ClassVar[int]
        key_type: Schema.FieldType
        value_type: Schema.FieldType
        def __init__(self, key_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ..., value_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ...) -> None: ...
    class TimeInfo(_message.Message):
        __slots__ = ["precision"]
        PRECISION_FIELD_NUMBER: _ClassVar[int]
        precision: int
        def __init__(self, precision: _Optional[int] = ...) -> None: ...
    class TimestampInfo(_message.Message):
        __slots__ = ["precision"]
        PRECISION_FIELD_NUMBER: _ClassVar[int]
        precision: int
        def __init__(self, precision: _Optional[int] = ...) -> None: ...
    class VarBinaryInfo(_message.Message):
        __slots__ = ["length"]
        LENGTH_FIELD_NUMBER: _ClassVar[int]
        length: int
        def __init__(self, length: _Optional[int] = ...) -> None: ...
    class VarCharInfo(_message.Message):
        __slots__ = ["length"]
        LENGTH_FIELD_NUMBER: _ClassVar[int]
        length: int
        def __init__(self, length: _Optional[int] = ...) -> None: ...
    class ZonedTimestampInfo(_message.Message):
        __slots__ = ["precision"]
        PRECISION_FIELD_NUMBER: _ClassVar[int]
        precision: int
        def __init__(self, precision: _Optional[int] = ...) -> None: ...
    BASIC_ARRAY: Schema.TypeName
    BIGINT: Schema.TypeName
    BINARY: Schema.TypeName
    BOOLEAN: Schema.TypeName
    CHAR: Schema.TypeName
    DATE: Schema.TypeName
    DECIMAL: Schema.TypeName
    DOUBLE: Schema.TypeName
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    FLOAT: Schema.TypeName
    INT: Schema.TypeName
    LOCAL_ZONED_TIMESTAMP: Schema.TypeName
    MAP: Schema.TypeName
    MULTISET: Schema.TypeName
    NULL: Schema.TypeName
    ROW: Schema.TypeName
    SMALLINT: Schema.TypeName
    TIME: Schema.TypeName
    TIMESTAMP: Schema.TypeName
    TINYINT: Schema.TypeName
    VARBINARY: Schema.TypeName
    VARCHAR: Schema.TypeName
    ZONED_TIMESTAMP: Schema.TypeName
    fields: _containers.RepeatedCompositeFieldContainer[Schema.Field]
    def __init__(self, fields: _Optional[_Iterable[_Union[Schema.Field, _Mapping]]] = ...) -> None: ...

class StateDescriptor(_message.Message):
    __slots__ = ["state_name", "state_ttl_config"]
    class StateTTLConfig(_message.Message):
        __slots__ = ["cleanup_strategies", "state_visibility", "ttl", "ttl_time_characteristic", "update_type"]
        class StateVisibility(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class TtlTimeCharacteristic(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class UpdateType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = []
        class CleanupStrategies(_message.Message):
            __slots__ = ["is_cleanup_in_background", "strategies"]
            class EmptyCleanupStrategy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class Strategies(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = []
            class IncrementalCleanupStrategy(_message.Message):
                __slots__ = ["cleanup_size", "run_cleanup_for_every_record"]
                CLEANUP_SIZE_FIELD_NUMBER: _ClassVar[int]
                RUN_CLEANUP_FOR_EVERY_RECORD_FIELD_NUMBER: _ClassVar[int]
                cleanup_size: int
                run_cleanup_for_every_record: bool
                def __init__(self, cleanup_size: _Optional[int] = ..., run_cleanup_for_every_record: bool = ...) -> None: ...
            class MapStrategiesEntry(_message.Message):
                __slots__ = ["empty_strategy", "incremental_cleanup_strategy", "rocksdb_compact_filter_cleanup_strategy", "strategy"]
                EMPTY_STRATEGY_FIELD_NUMBER: _ClassVar[int]
                INCREMENTAL_CLEANUP_STRATEGY_FIELD_NUMBER: _ClassVar[int]
                ROCKSDB_COMPACT_FILTER_CLEANUP_STRATEGY_FIELD_NUMBER: _ClassVar[int]
                STRATEGY_FIELD_NUMBER: _ClassVar[int]
                empty_strategy: StateDescriptor.StateTTLConfig.CleanupStrategies.EmptyCleanupStrategy
                incremental_cleanup_strategy: StateDescriptor.StateTTLConfig.CleanupStrategies.IncrementalCleanupStrategy
                rocksdb_compact_filter_cleanup_strategy: StateDescriptor.StateTTLConfig.CleanupStrategies.RocksdbCompactFilterCleanupStrategy
                strategy: StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies
                def __init__(self, strategy: _Optional[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies, str]] = ..., empty_strategy: _Optional[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies.EmptyCleanupStrategy, str]] = ..., incremental_cleanup_strategy: _Optional[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies.IncrementalCleanupStrategy, _Mapping]] = ..., rocksdb_compact_filter_cleanup_strategy: _Optional[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies.RocksdbCompactFilterCleanupStrategy, _Mapping]] = ...) -> None: ...
            class RocksdbCompactFilterCleanupStrategy(_message.Message):
                __slots__ = ["query_time_after_num_entries"]
                QUERY_TIME_AFTER_NUM_ENTRIES_FIELD_NUMBER: _ClassVar[int]
                query_time_after_num_entries: int
                def __init__(self, query_time_after_num_entries: _Optional[int] = ...) -> None: ...
            EMPTY_STRATEGY: StateDescriptor.StateTTLConfig.CleanupStrategies.EmptyCleanupStrategy
            FULL_STATE_SCAN_SNAPSHOT: StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies
            INCREMENTAL_CLEANUP: StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies
            IS_CLEANUP_IN_BACKGROUND_FIELD_NUMBER: _ClassVar[int]
            ROCKSDB_COMPACTION_FILTER: StateDescriptor.StateTTLConfig.CleanupStrategies.Strategies
            STRATEGIES_FIELD_NUMBER: _ClassVar[int]
            is_cleanup_in_background: bool
            strategies: _containers.RepeatedCompositeFieldContainer[StateDescriptor.StateTTLConfig.CleanupStrategies.MapStrategiesEntry]
            def __init__(self, is_cleanup_in_background: bool = ..., strategies: _Optional[_Iterable[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies.MapStrategiesEntry, _Mapping]]] = ...) -> None: ...
        CLEANUP_STRATEGIES_FIELD_NUMBER: _ClassVar[int]
        Disabled: StateDescriptor.StateTTLConfig.UpdateType
        NeverReturnExpired: StateDescriptor.StateTTLConfig.StateVisibility
        OnCreateAndWrite: StateDescriptor.StateTTLConfig.UpdateType
        OnReadAndWrite: StateDescriptor.StateTTLConfig.UpdateType
        ProcessingTime: StateDescriptor.StateTTLConfig.TtlTimeCharacteristic
        ReturnExpiredIfNotCleanedUp: StateDescriptor.StateTTLConfig.StateVisibility
        STATE_VISIBILITY_FIELD_NUMBER: _ClassVar[int]
        TTL_FIELD_NUMBER: _ClassVar[int]
        TTL_TIME_CHARACTERISTIC_FIELD_NUMBER: _ClassVar[int]
        UPDATE_TYPE_FIELD_NUMBER: _ClassVar[int]
        cleanup_strategies: StateDescriptor.StateTTLConfig.CleanupStrategies
        state_visibility: StateDescriptor.StateTTLConfig.StateVisibility
        ttl: int
        ttl_time_characteristic: StateDescriptor.StateTTLConfig.TtlTimeCharacteristic
        update_type: StateDescriptor.StateTTLConfig.UpdateType
        def __init__(self, update_type: _Optional[_Union[StateDescriptor.StateTTLConfig.UpdateType, str]] = ..., state_visibility: _Optional[_Union[StateDescriptor.StateTTLConfig.StateVisibility, str]] = ..., ttl_time_characteristic: _Optional[_Union[StateDescriptor.StateTTLConfig.TtlTimeCharacteristic, str]] = ..., ttl: _Optional[int] = ..., cleanup_strategies: _Optional[_Union[StateDescriptor.StateTTLConfig.CleanupStrategies, _Mapping]] = ...) -> None: ...
    STATE_NAME_FIELD_NUMBER: _ClassVar[int]
    STATE_TTL_CONFIG_FIELD_NUMBER: _ClassVar[int]
    state_name: str
    state_ttl_config: StateDescriptor.StateTTLConfig
    def __init__(self, state_name: _Optional[str] = ..., state_ttl_config: _Optional[_Union[StateDescriptor.StateTTLConfig, _Mapping]] = ...) -> None: ...

class TypeInfo(_message.Message):
    __slots__ = ["avro_type_info", "collection_element_type", "map_type_info", "row_type_info", "tuple_type_info", "type_name"]
    class TypeName(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class AvroTypeInfo(_message.Message):
        __slots__ = ["schema"]
        SCHEMA_FIELD_NUMBER: _ClassVar[int]
        schema: str
        def __init__(self, schema: _Optional[str] = ...) -> None: ...
    class MapTypeInfo(_message.Message):
        __slots__ = ["key_type", "value_type"]
        KEY_TYPE_FIELD_NUMBER: _ClassVar[int]
        VALUE_TYPE_FIELD_NUMBER: _ClassVar[int]
        key_type: TypeInfo
        value_type: TypeInfo
        def __init__(self, key_type: _Optional[_Union[TypeInfo, _Mapping]] = ..., value_type: _Optional[_Union[TypeInfo, _Mapping]] = ...) -> None: ...
    class RowTypeInfo(_message.Message):
        __slots__ = ["fields"]
        class Field(_message.Message):
            __slots__ = ["field_name", "field_type"]
            FIELD_NAME_FIELD_NUMBER: _ClassVar[int]
            FIELD_TYPE_FIELD_NUMBER: _ClassVar[int]
            field_name: str
            field_type: TypeInfo
            def __init__(self, field_name: _Optional[str] = ..., field_type: _Optional[_Union[TypeInfo, _Mapping]] = ...) -> None: ...
        FIELDS_FIELD_NUMBER: _ClassVar[int]
        fields: _containers.RepeatedCompositeFieldContainer[TypeInfo.RowTypeInfo.Field]
        def __init__(self, fields: _Optional[_Iterable[_Union[TypeInfo.RowTypeInfo.Field, _Mapping]]] = ...) -> None: ...
    class TupleTypeInfo(_message.Message):
        __slots__ = ["field_types"]
        FIELD_TYPES_FIELD_NUMBER: _ClassVar[int]
        field_types: _containers.RepeatedCompositeFieldContainer[TypeInfo]
        def __init__(self, field_types: _Optional[_Iterable[_Union[TypeInfo, _Mapping]]] = ...) -> None: ...
    AVRO: TypeInfo.TypeName
    AVRO_TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
    BASIC_ARRAY: TypeInfo.TypeName
    BIG_DEC: TypeInfo.TypeName
    BIG_INT: TypeInfo.TypeName
    BOOLEAN: TypeInfo.TypeName
    BYTE: TypeInfo.TypeName
    CHAR: TypeInfo.TypeName
    COLLECTION_ELEMENT_TYPE_FIELD_NUMBER: _ClassVar[int]
    DOUBLE: TypeInfo.TypeName
    FLOAT: TypeInfo.TypeName
    INSTANT: TypeInfo.TypeName
    INT: TypeInfo.TypeName
    LIST: TypeInfo.TypeName
    LOCAL_DATE: TypeInfo.TypeName
    LOCAL_DATETIME: TypeInfo.TypeName
    LOCAL_TIME: TypeInfo.TypeName
    LOCAL_ZONED_TIMESTAMP: TypeInfo.TypeName
    LONG: TypeInfo.TypeName
    MAP: TypeInfo.TypeName
    MAP_TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
    OBJECT_ARRAY: TypeInfo.TypeName
    PICKLED_BYTES: TypeInfo.TypeName
    PRIMITIVE_ARRAY: TypeInfo.TypeName
    ROW: TypeInfo.TypeName
    ROW_TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
    SHORT: TypeInfo.TypeName
    SQL_DATE: TypeInfo.TypeName
    SQL_TIME: TypeInfo.TypeName
    SQL_TIMESTAMP: TypeInfo.TypeName
    STRING: TypeInfo.TypeName
    TUPLE: TypeInfo.TypeName
    TUPLE_TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
    TYPE_NAME_FIELD_NUMBER: _ClassVar[int]
    avro_type_info: TypeInfo.AvroTypeInfo
    collection_element_type: TypeInfo
    map_type_info: TypeInfo.MapTypeInfo
    row_type_info: TypeInfo.RowTypeInfo
    tuple_type_info: TypeInfo.TupleTypeInfo
    type_name: TypeInfo.TypeName
    def __init__(self, type_name: _Optional[_Union[TypeInfo.TypeName, str]] = ..., collection_element_type: _Optional[_Union[TypeInfo, _Mapping]] = ..., row_type_info: _Optional[_Union[TypeInfo.RowTypeInfo, _Mapping]] = ..., tuple_type_info: _Optional[_Union[TypeInfo.TupleTypeInfo, _Mapping]] = ..., map_type_info: _Optional[_Union[TypeInfo.MapTypeInfo, _Mapping]] = ..., avro_type_info: _Optional[_Union[TypeInfo.AvroTypeInfo, _Mapping]] = ...) -> None: ...

class UserDefinedAggregateFunction(_message.Message):
    __slots__ = ["distinct", "filter_arg", "inputs", "payload", "specs", "takes_row_as_input"]
    class DataViewSpec(_message.Message):
        __slots__ = ["field_index", "list_view", "map_view", "name"]
        class ListView(_message.Message):
            __slots__ = ["element_type"]
            ELEMENT_TYPE_FIELD_NUMBER: _ClassVar[int]
            element_type: Schema.FieldType
            def __init__(self, element_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ...) -> None: ...
        class MapView(_message.Message):
            __slots__ = ["key_type", "value_type"]
            KEY_TYPE_FIELD_NUMBER: _ClassVar[int]
            VALUE_TYPE_FIELD_NUMBER: _ClassVar[int]
            key_type: Schema.FieldType
            value_type: Schema.FieldType
            def __init__(self, key_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ..., value_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ...) -> None: ...
        FIELD_INDEX_FIELD_NUMBER: _ClassVar[int]
        LIST_VIEW_FIELD_NUMBER: _ClassVar[int]
        MAP_VIEW_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        field_index: int
        list_view: UserDefinedAggregateFunction.DataViewSpec.ListView
        map_view: UserDefinedAggregateFunction.DataViewSpec.MapView
        name: str
        def __init__(self, name: _Optional[str] = ..., field_index: _Optional[int] = ..., list_view: _Optional[_Union[UserDefinedAggregateFunction.DataViewSpec.ListView, _Mapping]] = ..., map_view: _Optional[_Union[UserDefinedAggregateFunction.DataViewSpec.MapView, _Mapping]] = ...) -> None: ...
    DISTINCT_FIELD_NUMBER: _ClassVar[int]
    FILTER_ARG_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    SPECS_FIELD_NUMBER: _ClassVar[int]
    TAKES_ROW_AS_INPUT_FIELD_NUMBER: _ClassVar[int]
    distinct: bool
    filter_arg: int
    inputs: _containers.RepeatedCompositeFieldContainer[Input]
    payload: bytes
    specs: _containers.RepeatedCompositeFieldContainer[UserDefinedAggregateFunction.DataViewSpec]
    takes_row_as_input: bool
    def __init__(self, payload: _Optional[bytes] = ..., inputs: _Optional[_Iterable[_Union[Input, _Mapping]]] = ..., specs: _Optional[_Iterable[_Union[UserDefinedAggregateFunction.DataViewSpec, _Mapping]]] = ..., filter_arg: _Optional[int] = ..., distinct: bool = ..., takes_row_as_input: bool = ...) -> None: ...

class UserDefinedAggregateFunctions(_message.Message):
    __slots__ = ["count_star_inserted", "generate_update_before", "group_window", "grouping", "index_of_count_star", "job_parameters", "key_type", "map_state_read_cache_size", "map_state_write_cache_size", "metric_enabled", "profile_enabled", "state_cache_size", "state_cleaning_enabled", "udfs"]
    COUNT_STAR_INSERTED_FIELD_NUMBER: _ClassVar[int]
    GENERATE_UPDATE_BEFORE_FIELD_NUMBER: _ClassVar[int]
    GROUPING_FIELD_NUMBER: _ClassVar[int]
    GROUP_WINDOW_FIELD_NUMBER: _ClassVar[int]
    INDEX_OF_COUNT_STAR_FIELD_NUMBER: _ClassVar[int]
    JOB_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    KEY_TYPE_FIELD_NUMBER: _ClassVar[int]
    MAP_STATE_READ_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAP_STATE_WRITE_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    METRIC_ENABLED_FIELD_NUMBER: _ClassVar[int]
    PROFILE_ENABLED_FIELD_NUMBER: _ClassVar[int]
    STATE_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    STATE_CLEANING_ENABLED_FIELD_NUMBER: _ClassVar[int]
    UDFS_FIELD_NUMBER: _ClassVar[int]
    count_star_inserted: bool
    generate_update_before: bool
    group_window: GroupWindow
    grouping: _containers.RepeatedScalarFieldContainer[int]
    index_of_count_star: int
    job_parameters: _containers.RepeatedCompositeFieldContainer[JobParameter]
    key_type: Schema.FieldType
    map_state_read_cache_size: int
    map_state_write_cache_size: int
    metric_enabled: bool
    profile_enabled: bool
    state_cache_size: int
    state_cleaning_enabled: bool
    udfs: _containers.RepeatedCompositeFieldContainer[UserDefinedAggregateFunction]
    def __init__(self, udfs: _Optional[_Iterable[_Union[UserDefinedAggregateFunction, _Mapping]]] = ..., metric_enabled: bool = ..., grouping: _Optional[_Iterable[int]] = ..., generate_update_before: bool = ..., key_type: _Optional[_Union[Schema.FieldType, _Mapping]] = ..., index_of_count_star: _Optional[int] = ..., state_cleaning_enabled: bool = ..., state_cache_size: _Optional[int] = ..., map_state_read_cache_size: _Optional[int] = ..., map_state_write_cache_size: _Optional[int] = ..., count_star_inserted: bool = ..., group_window: _Optional[_Union[GroupWindow, _Mapping]] = ..., profile_enabled: bool = ..., job_parameters: _Optional[_Iterable[_Union[JobParameter, _Mapping]]] = ...) -> None: ...

class UserDefinedDataStreamFunction(_message.Message):
    __slots__ = ["function_type", "has_side_output", "key_type_info", "map_state_read_cache_size", "map_state_write_cache_size", "metric_enabled", "payload", "profile_enabled", "runtime_context", "state_cache_size"]
    class FunctionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    class RuntimeContext(_message.Message):
        __slots__ = ["attempt_number", "in_batch_execution_mode", "index_of_this_subtask", "job_parameters", "max_number_of_parallel_subtasks", "number_of_parallel_subtasks", "task_name", "task_name_with_subtasks"]
        ATTEMPT_NUMBER_FIELD_NUMBER: _ClassVar[int]
        INDEX_OF_THIS_SUBTASK_FIELD_NUMBER: _ClassVar[int]
        IN_BATCH_EXECUTION_MODE_FIELD_NUMBER: _ClassVar[int]
        JOB_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
        MAX_NUMBER_OF_PARALLEL_SUBTASKS_FIELD_NUMBER: _ClassVar[int]
        NUMBER_OF_PARALLEL_SUBTASKS_FIELD_NUMBER: _ClassVar[int]
        TASK_NAME_FIELD_NUMBER: _ClassVar[int]
        TASK_NAME_WITH_SUBTASKS_FIELD_NUMBER: _ClassVar[int]
        attempt_number: int
        in_batch_execution_mode: bool
        index_of_this_subtask: int
        job_parameters: _containers.RepeatedCompositeFieldContainer[JobParameter]
        max_number_of_parallel_subtasks: int
        number_of_parallel_subtasks: int
        task_name: str
        task_name_with_subtasks: str
        def __init__(self, task_name: _Optional[str] = ..., task_name_with_subtasks: _Optional[str] = ..., number_of_parallel_subtasks: _Optional[int] = ..., max_number_of_parallel_subtasks: _Optional[int] = ..., index_of_this_subtask: _Optional[int] = ..., attempt_number: _Optional[int] = ..., job_parameters: _Optional[_Iterable[_Union[JobParameter, _Mapping]]] = ..., in_batch_execution_mode: bool = ...) -> None: ...
    CO_BROADCAST_PROCESS: UserDefinedDataStreamFunction.FunctionType
    CO_PROCESS: UserDefinedDataStreamFunction.FunctionType
    FUNCTION_TYPE_FIELD_NUMBER: _ClassVar[int]
    HAS_SIDE_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    KEYED_CO_BROADCAST_PROCESS: UserDefinedDataStreamFunction.FunctionType
    KEYED_CO_PROCESS: UserDefinedDataStreamFunction.FunctionType
    KEYED_PROCESS: UserDefinedDataStreamFunction.FunctionType
    KEY_TYPE_INFO_FIELD_NUMBER: _ClassVar[int]
    MAP_STATE_READ_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAP_STATE_WRITE_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    METRIC_ENABLED_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    PROCESS: UserDefinedDataStreamFunction.FunctionType
    PROFILE_ENABLED_FIELD_NUMBER: _ClassVar[int]
    REVISE_OUTPUT: UserDefinedDataStreamFunction.FunctionType
    RUNTIME_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    STATE_CACHE_SIZE_FIELD_NUMBER: _ClassVar[int]
    WINDOW: UserDefinedDataStreamFunction.FunctionType
    function_type: UserDefinedDataStreamFunction.FunctionType
    has_side_output: bool
    key_type_info: TypeInfo
    map_state_read_cache_size: int
    map_state_write_cache_size: int
    metric_enabled: bool
    payload: bytes
    profile_enabled: bool
    runtime_context: UserDefinedDataStreamFunction.RuntimeContext
    state_cache_size: int
    def __init__(self, function_type: _Optional[_Union[UserDefinedDataStreamFunction.FunctionType, str]] = ..., runtime_context: _Optional[_Union[UserDefinedDataStreamFunction.RuntimeContext, _Mapping]] = ..., payload: _Optional[bytes] = ..., metric_enabled: bool = ..., key_type_info: _Optional[_Union[TypeInfo, _Mapping]] = ..., profile_enabled: bool = ..., has_side_output: bool = ..., state_cache_size: _Optional[int] = ..., map_state_read_cache_size: _Optional[int] = ..., map_state_write_cache_size: _Optional[int] = ...) -> None: ...

class UserDefinedFunction(_message.Message):
    __slots__ = ["inputs", "is_pandas_udf", "payload", "takes_row_as_input", "window_index"]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    IS_PANDAS_UDF_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    TAKES_ROW_AS_INPUT_FIELD_NUMBER: _ClassVar[int]
    WINDOW_INDEX_FIELD_NUMBER: _ClassVar[int]
    inputs: _containers.RepeatedCompositeFieldContainer[Input]
    is_pandas_udf: bool
    payload: bytes
    takes_row_as_input: bool
    window_index: int
    def __init__(self, payload: _Optional[bytes] = ..., inputs: _Optional[_Iterable[_Union[Input, _Mapping]]] = ..., window_index: _Optional[int] = ..., takes_row_as_input: bool = ..., is_pandas_udf: bool = ...) -> None: ...

class UserDefinedFunctions(_message.Message):
    __slots__ = ["job_parameters", "metric_enabled", "profile_enabled", "udfs", "windows"]
    JOB_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    METRIC_ENABLED_FIELD_NUMBER: _ClassVar[int]
    PROFILE_ENABLED_FIELD_NUMBER: _ClassVar[int]
    UDFS_FIELD_NUMBER: _ClassVar[int]
    WINDOWS_FIELD_NUMBER: _ClassVar[int]
    job_parameters: _containers.RepeatedCompositeFieldContainer[JobParameter]
    metric_enabled: bool
    profile_enabled: bool
    udfs: _containers.RepeatedCompositeFieldContainer[UserDefinedFunction]
    windows: _containers.RepeatedCompositeFieldContainer[OverWindow]
    def __init__(self, udfs: _Optional[_Iterable[_Union[UserDefinedFunction, _Mapping]]] = ..., metric_enabled: bool = ..., windows: _Optional[_Iterable[_Union[OverWindow, _Mapping]]] = ..., profile_enabled: bool = ..., job_parameters: _Optional[_Iterable[_Union[JobParameter, _Mapping]]] = ...) -> None: ...



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import base64
import collections
from abc import ABC, abstractmethod
from apache_beam.coders import coder_impl
from apache_beam.portability.api import beam_fn_api_pb2
from apache_beam.runners.worker.bundle_processor import SynchronousBagRuntimeState
from apache_beam.transforms import userstate
from enum import Enum
from functools import partial
from io import BytesIO
from typing import List, Tuple, Any, Dict, Collection, cast

from pyflink.datastream import ReduceFunction
from pyflink.datastream.functions import AggregateFunction
from pyflink.datastream.state import StateTtlConfig, MapStateDescriptor, OperatorStateStore
from pyflink.fn_execution.beam.beam_coders import FlinkCoder
from pyflink.fn_execution.coders import FieldCoder, MapCoder, from_type_info
from pyflink.fn_execution.flink_fn_execution_pb2 import StateDescriptor as pb2_StateDescriptor
from pyflink.fn_execution.internal_state import InternalKvState, N, InternalValueState, \
    InternalListState, InternalReducingState, InternalMergingState, InternalAggregatingState, \
    InternalMapState, InternalReadOnlyBroadcastState, InternalBroadcastState


class LRUCache(object):
    """
    A simple LRUCache implementation used to manage the internal runtime state.
    An internal runtime state is used to handle the data under a specific key of a "public" state.
    So the number of the internal runtime states may keep growing during the streaming task
    execution. To prevent the OOM caused by the unlimited growth, we introduce this LRUCache
    to evict the inactive internal runtime states.
    """

    def __init__(self, max_entries, default_entry):
        self._max_entries = max_entries
        self._default_entry = default_entry
        self._cache = collections.OrderedDict()
        self._on_evict = None

    def get(self, key):
        value = self._cache.pop(key, self._default_entry)
        if value != self._default_entry:
            # update the last access time
            self._cache[key] = value
        return value

    def put(self, key, value):
        self._cache[key] = value
        while len(self._cache) > self._max_entries:
            name, value = self._cache.popitem(last=False)
            if self._on_evict is not None:
                self._on_evict(name, value)

    def evict(self, key):
        value = self._cache.pop(key, self._default_entry)
        if self._on_evict is not None:
            self._on_evict(key, value)

    def evict_all(self):
        if self._on_evict is not None:
            for item in self._cache.items():
                self._on_evict(*item)
        self._cache.clear()

    def set_on_evict(self, func):
        self._on_evict = func

    def __len__(self):
        return len(self._cache)

    def __iter__(self):
        return iter(self._cache.values())

    def __contains__(self, key):
        return key in self._cache


class SynchronousKvRuntimeState(InternalKvState, ABC):
    """
    Base Class for partitioned State implementation.
    """

    def __init__(self, name: str, remote_state_backend: 'RemoteKeyedStateBackend'):
        self.name = name
        self._remote_state_backend = remote_state_backend
        self._internal_state = None
        self.namespace = None
        self._ttl_config = None
        self._cache_type = SynchronousKvRuntimeState.CacheType.ENABLE_READ_WRITE_CACHE

    def set_current_namespace(self, namespace: N) -> None:
        if namespace == self.namespace:
            return
        if self.namespace is not None:
            self._remote_state_backend.cache_internal_state(
                self._remote_state_backend._encoded_current_key, self)
        self.namespace = namespace
        self._internal_state = None

    def enable_time_to_live(self, ttl_config: StateTtlConfig):
        self._ttl_config = ttl_config
        if ttl_config.get_state_visibility() == StateTtlConfig.StateVisibility.NeverReturnExpired:
            self._cache_type = SynchronousKvRuntimeState.CacheType.DISABLE_CACHE
        elif ttl_config.get_update_type() == StateTtlConfig.UpdateType.OnReadAndWrite:
            self._cache_type = SynchronousKvRuntimeState.CacheType.ENABLE_WRITE_CACHE

        if self._cache_type != SynchronousKvRuntimeState.CacheType.ENABLE_READ_WRITE_CACHE:
            # disable read cache
            self._remote_state_backend._state_handler._state_cache._cache._max_entries = 0

    @abstractmethod
    def get_internal_state(self):
        pass

    class CacheType(Enum):
        DISABLE_CACHE = 0
        ENABLE_WRITE_CACHE = 1
        ENABLE_READ_WRITE_CACHE = 2


class SynchronousBagKvRuntimeState(SynchronousKvRuntimeState, ABC):
    """
    Base Class for State implementation backed by a :class:`SynchronousBagRuntimeState`.
    """
    def __init__(self, name: str, value_coder, remote_state_backend: 'RemoteKeyedStateBackend'):
        super(SynchronousBagKvRuntimeState, self).__init__(name, remote_state_backend)
        self._value_coder = value_coder

    def get_internal_state(self):
        if self._internal_state is None:
            self._internal_state = self._remote_state_backend._get_internal_bag_state(
                self.name, self.namespace, self._value_coder, self._ttl_config)
        return self._internal_state

    def _maybe_clear_write_cache(self):
        if self._cache_type == SynchronousKvRuntimeState.CacheType.DISABLE_CACHE or \
                self._remote_state_backend._state_cache_size <= 0:
            self._internal_state.commit()
            self._internal_state._cleared = False
            self._internal_state._added_elements = []


class SynchronousValueRuntimeState(SynchronousBagKvRuntimeState, InternalValueState):
    """
    The runtime ValueState implementation backed by a :class:`SynchronousBagRuntimeState`.
    """

    def __init__(self, name: str, value_coder, remote_state_backend: 'RemoteKeyedStateBackend'):
        super(SynchronousValueRuntimeState, self).__init__(name, value_coder, remote_state_backend)

    def value(self):
        for i in self.get_internal_state().read():
            return i
        return None

    def update(self, value) -> None:
        self.get_internal_state()
        self._internal_state.clear()
        self._internal_state.add(value)
        self._maybe_clear_write_cache()

    def clear(self) -> None:
        self.get_internal_state().clear()


class SynchronousMergingRuntimeState(SynchronousBagKvRuntimeState, InternalMergingState, ABC):
    """
    Base Class for MergingState implementation.
    """

    def __init__(self, name: str, value_coder, remote_state_backend: 'RemoteKeyedStateBackend'):
        super(SynchronousMergingRuntimeState, self).__init__(
            name, value_coder, remote_state_backend)

    def merge_namespaces(self, target: N, sources: Collection[N]) -> None:
        self._remote_state_backend.merge_namespaces(self, target, sources, self._ttl_config)


class SynchronousListRuntimeState(SynchronousMergingRuntimeState, InternalListState):
    """
    The runtime ListState implementation backed by a :class:`SynchronousBagRuntimeState`.
    """

    def __init__(self, name: str, value_coder, remote_state_backend: 'RemoteKeyedStateBackend'):
        super(SynchronousListRuntimeState, self).__init__(name, value_coder, remote_state_backend)

    def add(self, v):
        self.get_internal_state().add(v)
        self._maybe_clear_write_cache()

    def get(self):
        return self.get_internal_state().read()

    def add_all(self, values):
        self.get_internal_state()._added_elements.extend(values)
        self._maybe_clear_write_cache()

    def update(self, values):
        self.clear()
        self.add_all(values)
        self._maybe_clear_write_cache()

    def clear(self):
        self.get_internal_state().clear()


class SynchronousReducingRuntimeState(SynchronousMergingRuntimeState, InternalReducingState):
    """
    The runtime ReducingState implementation backed by a :class:`SynchronousBagRuntimeState`.
    """

    def __init__(self,
                 name: str,
                 value_coder,
                 remote_state_backend: 'RemoteKeyedStateBackend',
                 reduce_function: ReduceFunction):
        super(SynchronousReducingRuntimeState, self).__init__(
            name, value_coder, remote_state_backend)
        self._reduce_function = reduce_function

    def add(self, v):
        current_value = self.get()
        if current_value is None:
            self._internal_state.add(v)
        else:
            self._internal_state.clear()
            self._internal_state.add(self._reduce_function.reduce(current_value, v))
        self._maybe_clear_write_cache()

    def get(self):
        for i in self.get_internal_state().read():
            return i
        return None

    def clear(self):
        self.get_internal_state().clear()


class SynchronousAggregatingRuntimeState(SynchronousMergingRuntimeState, InternalAggregatingState):
    """
    The runtime AggregatingState implementation backed by a :class:`SynchronousBagRuntimeState`.
    """

    def __init__(self,
                 name: str,
                 value_coder,
                 remote_state_backend: 'RemoteKeyedStateBackend',
                 agg_function: AggregateFunction):
        super(SynchronousAggregatingRuntimeState, self).__init__(
            name, value_coder, remote_state_backend)
        self._agg_function = agg_function

    def add(self, v):
        if v is None:
            self.clear()
            return
        accumulator = self._get_accumulator()
        if accumulator is None:
            accumulator = self._agg_function.create_accumulator()
        accumulator = self._agg_function.add(v, accumulator)
        self._internal_state.clear()
        self._internal_state.add(accumulator)
        self._maybe_clear_write_cache()

    def get(self):
        accumulator = self._get_accumulator()
        if accumulator is None:
            return None
        else:
            return self._agg_function.get_result(accumulator)

    def _get_accumulator(self):
        for i in self.get_internal_state().read():
            return i
        return None

    def clear(self):
        self.get_internal_state().clear()


class CachedMapState(LRUCache):

    def __init__(self, max_entries):
        super(CachedMapState, self).__init__(max_entries, None)
        self._all_data_cached = False
        self._cached_keys = set()

        def on_evict(key, value):
            if value[0]:
                self._cached_keys.remove(key)
                self._all_data_cached = False

        self.set_on_evict(on_evict)

    def set_all_data_cached(self):
        self._all_data_cached = True

    def is_all_data_cached(self):
        return self._all_data_cached

    def put(self, key, exists_and_value):
        if exists_and_value[0]:
            self._cached_keys.add(key)
        super(CachedMapState, self).put(key, exists_and_value)

    def get_cached_keys(self):
        return self._cached_keys


class IterateType(Enum):
    ITEMS = 0
    KEYS = 1
    VALUES = 2


class IteratorToken(Enum):
    """
    The token indicates the status of current underlying iterator. It can also be a UUID,
    which represents an iterator on the Java side.
    """
    NOT_START = 0
    FINISHED = 1


def create_cache_iterator(cache_dict, iterate_type, iterated_keys=None):
    if iterated_keys is None:
        iterated_keys = []
    if iterate_type == IterateType.KEYS:
        for key, (exists, value) in cache_dict.items():
            if not exists or key in iterated_keys:
                continue
            yield key, key
    elif iterate_type == IterateType.VALUES:
        for key, (exists, value) in cache_dict.items():
            if not exists or key in iterated_keys:
                continue
            yield key, value
    elif iterate_type == IterateType.ITEMS:
        for key, (exists, value) in cache_dict.items():
            if not exists or key in iterated_keys:
                continue
            yield key, (key, value)
    else:
        raise Exception("Unsupported iterate type: %s" % iterate_type)


class CachingMapStateHandler(object):
    # GET request flags
    GET_FLAG = 0
    ITERATE_FLAG = 1
    CHECK_EMPTY_FLAG = 2
    # GET response flags
    EXIST_FLAG = 0
    IS_NONE_FLAG = 1
    NOT_EXIST_FLAG = 2
    IS_EMPTY_FLAG = 3
    NOT_EMPTY_FLAG = 4
    # APPEND request flags
    DELETE = 0
    SET_NONE = 1
    SET_VALUE = 2

    def __init__(self, caching_state_handler, max_cached_map_key_entries):
        self._state_cache = caching_state_handler._state_cache
        self._underlying = caching_state_handler._underlying
        self._context = caching_state_handler._context
        self._max_cached_map_key_entries = max_cached_map_key_entries
        self._cached_iterator_num = 0

    def _get_cache_token(self):
        if not self._state_cache.is_cache_enabled():
            return None
        if self._context.user_state_cache_token:
            return self._context.user_state_cache_token
        else:
            return self._context.bundle_cache_token

    def blocking_get(self, state_key, map_key, map_key_encoder, map_value_decoder):
        cache_token = self._get_cache_token()
        if not cache_token:
            # cache disabled / no cache token, request from remote directly
            return self._get_raw(state_key, map_key, map_key_encoder, map_value_decoder)

        # lookup cache first
        cache_state_key = self._convert_to_cache_key(state_key)
        cached_map_state = self._state_cache.peek((cache_state_key, cache_token))
        if cached_map_state is None:
            # request from remote
            exists, value = self._get_raw(state_key, map_key, map_key_encoder, map_value_decoder)
            cached_map_state = CachedMapState(self._max_cached_map_key_entries)
            cached_map_state.put(map_key, (exists, value))
            self._state_cache.put((cache_state_key, cache_token), cached_map_state)
            return exists, value
        else:
            cached_value = cached_map_state.get(map_key)
            if cached_value is None:
                if cached_map_state.is_all_data_cached():
                    return False, None

                # request from remote
                exists, value = self._get_raw(
                    state_key, map_key, map_key_encoder, map_value_decoder)
                cached_map_state.put(map_key, (exists, value))
                return exists, value
            else:
                return cached_value

    def lazy_iterator(self, state_key, iterate_type, map_key_decoder, map_value_decoder,
                      iterated_keys):
        cache_token = self._get_cache_token()
        if cache_token:
            # check if the data in the read cache can be used
            cache_state_key = self._convert_to_cache_key(state_key)
            cached_map_state = self._state_cache.peek((cache_state_key, cache_token))
            if cached_map_state and cached_map_state.is_all_data_cached():
                return create_cache_iterator(
                    cached_map_state._cache, iterate_type, iterated_keys)

        # request from remote
        last_iterator_token = IteratorToken.NOT_START
        current_batch, iterator_token = self._iterate_raw(
            state_key, iterate_type,
            last_iterator_token,
            map_key_decoder,
            map_value_decoder)

        if cache_token and \
                iterator_token == IteratorToken.FINISHED and \
                iterate_type != IterateType.KEYS and \
                self._max_cached_map_key_entries >= len(current_batch):
            # Special case: all the data of the map state is contained in current batch,
            # and can be stored in the cached map state.
            cached_map_state = CachedMapState(self._max_cached_map_key_entries)
            cache_state_key = self._convert_to_cache_key(state_key)
            for key, value in current_batch.items():
                cached_map_state.put(key, (True, value))
            cached_map_state.set_all_data_cached()
            self._state_cache.put((cache_state_key, cache_token), cached_map_state)

        return self._lazy_remote_iterator(
            state_key,
            iterate_type,
            map_key_decoder,
            map_value_decoder,
            iterated_keys,
            iterator_token,
            current_batch)

    def _lazy_remote_iterator(
            self,
            state_key,
            iterate_type,
            map_key_decoder,
            map_value_decoder,
            iterated_keys,
            iterator_token,
            current_batch):
        if iterate_type == IterateType.KEYS:
            while True:
                for key in current_batch:
                    if key in iterated_keys:
                        continue
                    yield key, key
                if iterator_token == IteratorToken.FINISHED:
                    break
                current_batch, iterator_token = self._iterate_raw(
                    state_key,
                    iterate_type,
                    iterator_token,
                    map_key_decoder,
                    map_value_decoder)
        elif iterate_type == IterateType.VALUES:
            while True:
                for key, value in current_batch.items():
                    if key in iterated_keys:
                        continue
                    yield key, value
                if iterator_token == IteratorToken.FINISHED:
                    break
                current_batch, iterator_token = self._iterate_raw(
                    state_key,
                    iterate_type,
                    iterator_token,
                    map_key_decoder,
                    map_value_decoder)
        elif iterate_type == IterateType.ITEMS:
            while True:
                for key, value in current_batch.items():
                    if key in iterated_keys:
                        continue
                    yield key, (key, value)
                if iterator_token == IteratorToken.FINISHED:
                    break
                current_batch, iterator_token = self._iterate_raw(
                    state_key,
                    iterate_type,
                    iterator_token,
                    map_key_decoder,
                    map_value_decoder)
        else:
            raise Exception("Unsupported iterate type: %s" % iterate_type)

    def extend(self, state_key, items: List[Tuple[int, Any, Any]],
               map_key_encoder, map_value_encoder):
        cache_token = self._get_cache_token()
        if cache_token:
            # Cache lookup
            cache_state_key = self._convert_to_cache_key(state_key)
            cached_map_state = self._state_cache.peek((cache_state_key, cache_token))
            if cached_map_state is None:
                cached_map_state = CachedMapState(self._max_cached_map_key_entries)
                self._state_cache.put((cache_state_key, cache_token), cached_map_state)
            for request_flag, map_key, map_value in items:
                if request_flag == self.DELETE:
                    cached_map_state.put(map_key, (False, None))
                elif request_flag == self.SET_NONE:
                    cached_map_state.put(map_key, (True, None))
                elif request_flag == self.SET_VALUE:
                    cached_map_state.put(map_key, (True, map_value))
                else:
                    raise Exception("Unknown flag: " + str(request_flag))
        return self._append_raw(
            state_key,
            items,
            map_key_encoder,
            map_value_encoder)

    def check_empty(self, state_key):
        cache_token = self._get_cache_token()
        if cache_token:
            # Cache lookup
            cache_state_key = self._convert_to_cache_key(state_key)
            cached_map_state = self._state_cache.peek((cache_state_key, cache_token))
            if cached_map_state is not None:
                if cached_map_state.is_all_data_cached() and \
                        len(cached_map_state.get_cached_keys()) == 0:
                    return True
                elif len(cached_map_state.get_cached_keys()) > 0:
                    return False
        return self._check_empty_raw(state_key)

    def clear(self, state_key):
        self.clear_read_cache(state_key)
        return self._underlying.clear(state_key)

    def clear_read_cache(self, state_key):
        cache_token = self._get_cache_token()
        if cache_token:
            cache_key = self._convert_to_cache_key(state_key)
            self._state_cache.invalidate((cache_key, cache_token))

    def get_cached_iterators_num(self):
        return self._cached_iterator_num

    def _inc_cached_iterators_num(self):
        self._cached_iterator_num += 1

    def _dec_cached_iterators_num(self):
        self._cached_iterator_num -= 1

    def reset_cached_iterators_num(self):
        self._cached_iterator_num = 0

    def _check_empty_raw(self, state_key):
        output_stream = coder_impl.create_OutputStream()
        output_stream.write_byte(self.CHECK_EMPTY_FLAG)
        continuation_token = output_stream.get()
        data, response_token = self._underlying.get_raw(state_key, continuation_token)
        if data[0] == self.IS_EMPTY_FLAG:
            return True
        elif data[0] == self.NOT_EMPTY_FLAG:
            return False
        else:
            raise Exception("Unknown response flag: " + str(data[0]))

    def _get_raw(self, state_key, map_key, map_key_encoder, map_value_decoder):
        output_stream = coder_impl.create_OutputStream()
        output_stream.write_byte(self.GET_FLAG)
        map_key_encoder(map_key, output_stream)
        continuation_token = output_stream.get()
        data, response_token = self._underlying.get_raw(state_key, continuation_token)
        input_stream = coder_impl.create_InputStream(data)
        result_flag = input_stream.read_byte()
        if result_flag == self.EXIST_FLAG:
            return True, map_value_decoder(input_stream)
        elif result_flag == self.IS_NONE_FLAG:
            return True, None
        elif result_flag == self.NOT_EXIST_FLAG:
            return False, None
        else:
            raise Exception("Unknown response flag: " + str(result_flag))

    def _iterate_raw(self, state_key, iterate_type, iterator_token,
                     map_key_decoder, map_value_decoder):
        output_stream = coder_impl.create_OutputStream()
        output_stream.write_byte(self.ITERATE_FLAG)
        output_stream.write_byte(iterate_type.value)
        if not isinstance(iterator_token, IteratorToken):
            # The iterator token represents a Java iterator
            output_stream.write_bigendian_int32(len(iterator_token))
            output_stream.write(iterator_token)
        else:
            output_stream.write_bigendian_int32(0)
        continuation_token = output_stream.get()
        data, response_token = self._underlying.get_raw(state_key, continuation_token)
        if len(response_token) != 0:
            # The new iterator token is an UUID which represents a cached iterator at Java
            # side.
            new_iterator_token = response_token
            if iterator_token == IteratorToken.NOT_START:
                # This is the first request but not the last request of current state.
                # It means there is a new iterator has been created and cached at Java side.
                self._inc_cached_iterators_num()
        else:
            new_iterator_token = IteratorToken.FINISHED
            if iterator_token != IteratorToken.NOT_START:
                # This is not the first request but the last request of current state.
                # It means the cached iterator created at Java side has been removed as
                # current iteration has finished.
                self._dec_cached_iterators_num()
        input_stream = coder_impl.create_InputStream(data)
        if iterate_type == IterateType.ITEMS or iterate_type == IterateType.VALUES:
            # decode both key and value
            current_batch = {}
            while input_stream.size() > 0:
                key = map_key_decoder(input_stream)
                is_not_none = input_stream.read_byte()
                if is_not_none:
                    value = map_value_decoder(input_stream)
                else:
                    value = None
                current_batch[key] = value
        else:
            # only decode key
            current_batch = []
            while input_stream.size() > 0:
                key = map_key_decoder(input_stream)
                current_batch.append(key)
        return current_batch, new_iterator_token

    def _append_raw(self, state_key, items, map_key_encoder, map_value_encoder):
        output_stream = coder_impl.create_OutputStream()
        output_stream.write_bigendian_int32(len(items))
        for request_flag, map_key, map_value in items:
            output_stream.write_byte(request_flag)
            # Not all the coder impls will serialize the length of bytes when we set the "nested"
            # param to "True", so we need to encode the length of bytes manually.
            tmp_out = coder_impl.create_OutputStream()
            map_key_encoder(map_key, tmp_out)
            serialized_data = tmp_out.get()
            output_stream.write_bigendian_int32(len(serialized_data))
            output_stream.write(serialized_data)
            if request_flag == self.SET_VALUE:
                tmp_out = coder_impl.create_OutputStream()
                map_value_encoder(map_value, tmp_out)
                serialized_data = tmp_out.get()
                output_stream.write_bigendian_int32(len(serialized_data))
                output_stream.write(serialized_data)
        return self._underlying.append_raw(state_key, output_stream.get())

    @staticmethod
    def _convert_to_cache_key(state_key):
        return state_key.SerializeToString()


class RemovableConcatIterator(collections.abc.Iterator):

    def __init__(self, internal_map_state, first, second):
        self._first = first
        self._second = second
        self._first_not_finished = True
        self._internal_map_state = internal_map_state
        self._mod_count = self._internal_map_state._mod_count
        self._last_key = None

    def __next__(self):
        self._check_modification()
        if self._first_not_finished:
            try:
                self._last_key, element = next(self._first)
                return element
            except StopIteration:
                self._first_not_finished = False
                return self.__next__()
        else:
            self._last_key, element = next(self._second)
            return element

    def remove(self):
        """
        Remove the last element returned by this iterator.
        """
        if self._last_key is None:
            raise Exception("You need to call the '__next__' method before calling "
                            "this method.")
        self._check_modification()
        # Bypass the 'remove' method of the map state to avoid triggering the commit of the write
        # cache.
        if self._internal_map_state._cleared:
            del self._internal_map_state._write_cache[self._last_key]
            if len(self._internal_map_state._write_cache) == 0:
                self._internal_map_state._is_empty = True
        else:
            self._internal_map_state._write_cache[self._last_key] = (False, None)
        self._mod_count += 1
        self._internal_map_state._mod_count += 1
        self._last_key = None

    def _check_modification(self):
        if self._mod_count != self._internal_map_state._mod_count:
            raise Exception("Concurrent modification detected. "
                            "You can not modify the map state when iterating it except using the "
                            "'remove' method of this iterator.")


class InternalSynchronousMapRuntimeState(object):

    def __init__(self,
                 map_state_handler: CachingMapStateHandler,
                 state_key,
                 map_key_coder,
                 map_value_coder,
                 max_write_cache_entries):
        self._map_state_handler = map_state_handler
        self._state_key = state_key
        self._map_key_coder = map_key_coder
        if isinstance(map_key_coder, FieldCoder):
            map_key_coder_impl = FlinkCoder(map_key_coder).get_impl()
        else:
            map_key_coder_impl = map_key_coder.get_impl()
        self._map_key_encoder, self._map_key_decoder = \
            self._get_encoder_and_decoder(map_key_coder_impl)
        self._map_value_coder = map_value_coder
        if isinstance(map_value_coder, FieldCoder):
            map_value_coder_impl = FlinkCoder(map_value_coder).get_impl()
        else:
            map_value_coder_impl = map_value_coder.get_impl()
        self._map_value_encoder, self._map_value_decoder = \
            self._get_encoder_and_decoder(map_value_coder_impl)
        self._write_cache = dict()
        self._max_write_cache_entries = max_write_cache_entries
        self._is_empty = None
        self._cleared = False
        self._mod_count = 0

    def get(self, map_key):
        if self._is_empty:
            return None
        if map_key in self._write_cache:
            exists, value = self._write_cache[map_key]
            if exists:
                return value
            else:
                return None
        if self._cleared:
            return None
        exists, value = self._map_state_handler.blocking_get(
            self._state_key, map_key, self._map_key_encoder, self._map_value_decoder)
        if exists:
            return value
        else:
            return None

    def put(self, map_key, map_value):
        self._write_cache[map_key] = (True, map_value)
        self._is_empty = False
        self._mod_count += 1
        if len(self._write_cache) >= self._max_write_cache_entries:
            self.commit()

    def put_all(self, dict_value):
        for map_key, map_value in dict_value:
            self._write_cache[map_key] = (True, map_value)
        self._is_empty = False
        self._mod_count += 1
        if len(self._write_cache) >= self._max_write_cache_entries:
            self.commit()

    def remove(self, map_key):
        if self._is_empty:
            return
        if self._cleared:
            del self._write_cache[map_key]
            if len(self._write_cache) == 0:
                self._is_empty = True
        else:
            self._write_cache[map_key] = (False, None)
            self._is_empty = None
        self._mod_count += 1
        if len(self._write_cache) >= self._max_write_cache_entries:
            self.commit()

    def contains(self, map_key):
        if self._is_empty:
            return False
        if self.get(map_key) is None:
            return False
        else:
            return True

    def is_empty(self):
        if self._is_empty is None:
            if len(self._write_cache) > 0:
                self.commit()
            self._is_empty = self._map_state_handler.check_empty(self._state_key)
        return self._is_empty

    def clear(self):
        self._cleared = True
        self._is_empty = True
        self._mod_count += 1
        self._write_cache.clear()

    def items(self):
        return RemovableConcatIterator(
            self,
            self.write_cache_iterator(IterateType.ITEMS),
            self.remote_data_iterator(IterateType.ITEMS))

    def keys(self):
        return RemovableConcatIterator(
            self,
            self.write_cache_iterator(IterateType.KEYS),
            self.remote_data_iterator(IterateType.KEYS))

    def values(self):
        return RemovableConcatIterator(
            self,
            self.write_cache_iterator(IterateType.VALUES),
            self.remote_data_iterator(IterateType.VALUES))

    def commit(self):
        to_await = None
        if self._cleared:
            to_await = self._map_state_handler.clear(self._state_key)
        if self._write_cache:
            append_items = []
            for map_key, (exists, value) in self._write_cache.items():
                if exists:
                    if value is not None:
                        append_items.append(
                            (CachingMapStateHandler.SET_VALUE, map_key, value))
                    else:
                        append_items.append((CachingMapStateHandler.SET_NONE, map_key, None))
                else:
                    append_items.append((CachingMapStateHandler.DELETE, map_key, None))
            self._write_cache.clear()
            to_await = self._map_state_handler.extend(
                self._state_key, append_items, self._map_key_encoder, self._map_value_encoder)
        if to_await:
            to_await.get()
        self._write_cache.clear()
        self._cleared = False
        self._mod_count += 1

    def write_cache_iterator(self, iterate_type):
        return create_cache_iterator(self._write_cache, iterate_type)

    def remote_data_iterator(self, iterate_type):
        if self._cleared or self._is_empty:
            return iter([])
        else:
            return self._map_state_handler.lazy_iterator(
                self._state_key,
                iterate_type,
                self._map_key_decoder,
                self._map_value_decoder,
                self._write_cache)

    @staticmethod
    def _get_encoder_and_decoder(coder):
        encoder = partial(coder.encode_to_stream, nested=True)
        decoder = partial(coder.decode_from_stream, nested=True)
        return encoder, decoder


class SynchronousMapRuntimeState(SynchronousKvRuntimeState, InternalMapState):

    def __init__(self,
                 name: str,
                 map_key_coder,
                 map_value_coder,
                 remote_state_backend: 'RemoteKeyedStateBackend'):
        super(SynchronousMapRuntimeState, self).__init__(name, remote_state_backend)
        self._map_key_coder = map_key_coder
        self._map_value_coder = map_value_coder

    def get_internal_state(self):
        if self._internal_state is None:
            self._internal_state = self._remote_state_backend._get_internal_map_state(
                self.name,
                self.namespace,
                self._map_key_coder,
                self._map_value_coder,
                self._ttl_config,
                self._cache_type)
        return self._internal_state

    def get(self, key):
        return self.get_internal_state().get(key)

    def put(self, key, value):
        self.get_internal_state().put(key, value)

    def put_all(self, dict_value):
        self.get_internal_state().put_all(dict_value)

    def remove(self, key):
        self.get_internal_state().remove(key)

    def contains(self, key):
        return self.get_internal_state().contains(key)

    def items(self):
        return self.get_internal_state().items()

    def keys(self):
        return self.get_internal_state().keys()

    def values(self):
        return self.get_internal_state().values()

    def is_empty(self):
        return self.get_internal_state().is_empty()

    def clear(self):
        self.get_internal_state().clear()


class RemoteKeyedStateBackend(object):
    """
    A keyed state backend provides methods for managing keyed state.
    """

    MERGE_NAMESAPCES_MARK = "merge_namespaces"

    def __init__(self,
                 state_handler,
                 key_coder,
                 namespace_coder,
                 state_cache_size,
                 map_state_read_cache_size,
                 map_state_write_cache_size):
        self._state_handler = state_handler
        self._map_state_handler = CachingMapStateHandler(
            state_handler, map_state_read_cache_size)
        self._key_coder_impl = key_coder.get_impl()
        self.namespace_coder = namespace_coder
        if namespace_coder:
            self._namespace_coder_impl = namespace_coder.get_impl()
        else:
            self._namespace_coder_impl = None
        self._state_cache_size = state_cache_size
        self._map_state_write_cache_size = map_state_write_cache_size
        self._all_states = {}  # type: Dict[str, SynchronousKvRuntimeState]
        self._internal_state_cache = LRUCache(self._state_cache_size, None)
        self._internal_state_cache.set_on_evict(
            lambda key, value: self.commit_internal_state(value))
        self._current_key = None
        self._encoded_current_key = None
        self._clear_iterator_mark = beam_fn_api_pb2.StateKey(
            multimap_side_input=beam_fn_api_pb2.StateKey.MultimapSideInput(
                transform_id="clear_iterators",
                side_input_id="clear_iterators",
                key=self._encoded_current_key))

    def get_list_state(self, name, element_coder, ttl_config=None):
        return self._wrap_internal_bag_state(
            name,
            element_coder,
            SynchronousListRuntimeState,
            SynchronousListRuntimeState,
            ttl_config)

    def get_value_state(self, name, value_coder, ttl_config=None):
        return self._wrap_internal_bag_state(
            name,
            value_coder,
            SynchronousValueRuntimeState,
            SynchronousValueRuntimeState,
            ttl_config)

    def get_map_state(self, name, map_key_coder, map_value_coder, ttl_config=None):
        if name in self._all_states:
            self.validate_map_state(name, map_key_coder, map_value_coder)
            return self._all_states[name]
        map_state = SynchronousMapRuntimeState(name, map_key_coder, map_value_coder, self)
        if ttl_config is not None:
            map_state.enable_time_to_live(ttl_config)
        self._all_states[name] = map_state
        return map_state

    def get_reducing_state(self, name, coder, reduce_function, ttl_config=None):
        return self._wrap_internal_bag_state(
            name,
            coder,
            SynchronousReducingRuntimeState,
            partial(SynchronousReducingRuntimeState, reduce_function=reduce_function),
            ttl_config)

    def get_aggregating_state(self, name, coder, agg_function, ttl_config=None):
        return self._wrap_internal_bag_state(
            name,
            coder,
            SynchronousAggregatingRuntimeState,
            partial(SynchronousAggregatingRuntimeState, agg_function=agg_function),
            ttl_config)

    def validate_state(self, name, coder, expected_type):
        if name in self._all_states:
            state = self._all_states[name]
            if not isinstance(state, expected_type):
                raise Exception("The state name '%s' is already in use and not a %s."
                                % (name, expected_type))
            if state._value_coder != coder:
                raise Exception("State name corrupted: %s" % name)

    def validate_map_state(self, name, map_key_coder, map_value_coder):
        if name in self._all_states:
            state = self._all_states[name]
            if not isinstance(state, SynchronousMapRuntimeState):
                raise Exception("The state name '%s' is already in use and not a map state."
                                % name)
            if state._map_key_coder != map_key_coder or \
                    state._map_value_coder != map_value_coder:
                raise Exception("State name corrupted: %s" % name)

    def _wrap_internal_bag_state(
            self, name, element_coder, wrapper_type, wrap_method, ttl_config):
        if name in self._all_states:
            self.validate_state(name, element_coder, wrapper_type)
            return self._all_states[name]
        wrapped_state = wrap_method(name, element_coder, self)
        if ttl_config is not None:
            wrapped_state.enable_time_to_live(ttl_config)
        self._all_states[name] = wrapped_state
        return wrapped_state

    def _get_internal_bag_state(self, name, namespace, element_coder, ttl_config):
        encoded_namespace = self._encode_namespace(namespace)
        cached_state = self._internal_state_cache.get(
            (name, self._encoded_current_key, encoded_namespace))
        if cached_state is not None:
            return cached_state
        # The created internal state would not be put into the internal state cache
        # at once. The internal state cache is only updated when the current key changes.
        # The reason is that the state cache size may be smaller that the count of activated
        # state (i.e. the state with current key).
        if isinstance(element_coder, FieldCoder):
            element_coder = FlinkCoder(element_coder)
        state_spec = userstate.BagStateSpec(name, element_coder)
        internal_state = self._create_bag_state(state_spec, encoded_namespace, ttl_config)
        return internal_state

    def _get_internal_map_state(
            self, name, namespace, map_key_coder, map_value_coder, ttl_config, cache_type):
        encoded_namespace = self._encode_namespace(namespace)
        cached_state = self._internal_state_cache.get(
            (name, self._encoded_current_key, encoded_namespace))
        if cached_state is not None:
            return cached_state
        internal_map_state = self._create_internal_map_state(
            name, encoded_namespace, map_key_coder, map_value_coder, ttl_config, cache_type)
        return internal_map_state

    def _create_bag_state(self, state_spec: userstate.StateSpec, encoded_namespace, ttl_config) \
            -> userstate.AccumulatingRuntimeState:
        if isinstance(state_spec, userstate.BagStateSpec):
            bag_state = SynchronousBagRuntimeState(
                self._state_handler,
                state_key=self.get_bag_state_key(
                    state_spec.name, self._encoded_current_key, encoded_namespace, ttl_config),
                value_coder=state_spec.coder)
            return bag_state
        else:
            raise NotImplementedError(state_spec)

    def _create_internal_map_state(
            self, name, encoded_namespace, map_key_coder, map_value_coder, ttl_config, cache_type):
        # Currently the `beam_fn_api.proto` does not support MapState, so we use the
        # the `MultimapSideInput` message to mark the state as a MapState for now.
        state_proto = pb2_StateDescriptor()
        state_proto.state_name = name
        if ttl_config is not None:
            state_proto.state_ttl_config.CopyFrom(ttl_config._to_proto())
        state_key = beam_fn_api_pb2.StateKey(
            multimap_side_input=beam_fn_api_pb2.StateKey.MultimapSideInput(
                transform_id="",
                window=encoded_namespace,
                side_input_id=base64.b64encode(state_proto.SerializeToString()),
                key=self._encoded_current_key))
        if cache_type == SynchronousKvRuntimeState.CacheType.DISABLE_CACHE:
            write_cache_size = 0
        else:
            write_cache_size = self._map_state_write_cache_size
        return InternalSynchronousMapRuntimeState(
            self._map_state_handler,
            state_key,
            map_key_coder,
            map_value_coder,
            write_cache_size)

    def _encode_namespace(self, namespace):
        if namespace is not None:
            encoded_namespace = self._namespace_coder_impl.encode(namespace)
        else:
            encoded_namespace = b''
        return encoded_namespace

    def cache_internal_state(self, encoded_key, internal_kv_state: SynchronousKvRuntimeState):
        encoded_old_namespace = self._encode_namespace(internal_kv_state.namespace)
        self._internal_state_cache.put(
            (internal_kv_state.name, encoded_key, encoded_old_namespace),
            internal_kv_state.get_internal_state())

    def set_current_key(self, key):
        if key == self._current_key:
            return
        encoded_old_key = self._encoded_current_key
        for state_name, state_obj in self._all_states.items():
            if self._state_cache_size > 0:
                # cache old internal state
                self.cache_internal_state(encoded_old_key, state_obj)
            state_obj.namespace = None
            state_obj._internal_state = None
        self._current_key = key
        self._encoded_current_key = self._key_coder_impl.encode(self._current_key)

    def get_current_key(self):
        return self._current_key

    def commit(self):
        for internal_state in self._internal_state_cache:
            self.commit_internal_state(internal_state)
        for name, state in self._all_states.items():
            if (name, self._encoded_current_key, self._encode_namespace(state.namespace)) \
                    not in self._internal_state_cache:
                self.commit_internal_state(state._internal_state)

    def clear_cached_iterators(self):
        if self._map_state_handler.get_cached_iterators_num() > 0:
            self._clear_iterator_mark.multimap_side_input.key = self._encoded_current_key
            self._map_state_handler.clear(self._clear_iterator_mark)

    def merge_namespaces(self, state: SynchronousMergingRuntimeState, target, sources, ttl_config):
        for source in sources:
            state.set_current_namespace(source)
            self.commit_internal_state(state.get_internal_state())
        state.set_current_namespace(target)
        self.commit_internal_state(state.get_internal_state())
        encoded_target_namespace = self._encode_namespace(target)
        encoded_namespaces = [self._encode_namespace(source) for source in sources]
        self.clear_state_cache(state, [encoded_target_namespace] + encoded_namespaces)

        state_key = self.get_bag_state_key(
            state.name, self._encoded_current_key, encoded_target_namespace, ttl_config)
        state_key.bag_user_state.transform_id = self.MERGE_NAMESAPCES_MARK

        encoded_namespaces_writer = BytesIO()
        encoded_namespaces_writer.write(len(sources).to_bytes(4, 'big'))
        for encoded_namespace in encoded_namespaces:
            encoded_namespaces_writer.write(encoded_namespace)
        sources_bytes = encoded_namespaces_writer.getvalue()
        to_await = self._map_state_handler._underlying.append_raw(state_key, sources_bytes)
        if to_await:
            to_await.get()

    def clear_state_cache(self, state: SynchronousMergingRuntimeState, encoded_namespaces):
        name = state.name
        for encoded_namespace in encoded_namespaces:
            if (name, self._encoded_current_key, encoded_namespace) in self._internal_state_cache:
                # commit and clear the write cache
                self._internal_state_cache.evict(
                    (name, self._encoded_current_key, encoded_namespace))
                # currently all the SynchronousMergingRuntimeState is based on bag state
                state_key = self.get_bag_state_key(
                    name, self._encoded_current_key, encoded_namespace, None)
                # clear the read cache, the read cache is shared between map state handler and bag
                # state handler. So we can use the map state handler instead.
                self._map_state_handler.clear_read_cache(state_key)

    def get_bag_state_key(self, name, encoded_key, encoded_namespace, ttl_config):
        from pyflink.fn_execution.flink_fn_execution_pb2 import StateDescriptor
        state_proto = StateDescriptor()
        state_proto.state_name = name
        if ttl_config is not None:
            state_proto.state_ttl_config.CopyFrom(ttl_config._to_proto())
        return beam_fn_api_pb2.StateKey(
            bag_user_state=beam_fn_api_pb2.StateKey.BagUserState(
                transform_id="",
                window=encoded_namespace,
                user_state_id=base64.b64encode(state_proto.SerializeToString()),
                key=encoded_key))

    @staticmethod
    def commit_internal_state(internal_state):
        if internal_state is not None:
            internal_state.commit()
        # reset the status of the internal state to reuse the object cross bundle
        if isinstance(internal_state, SynchronousBagRuntimeState):
            internal_state._cleared = False
            internal_state._added_elements = []


class SynchronousReadOnlyBroadcastRuntimeState(InternalReadOnlyBroadcastState):
    def __init__(self, name: str, internal_map_state: "InternalSynchronousMapRuntimeState"):
        self._name = name
        self._internal_map_state = internal_map_state

    def get(self, key):
        return self._internal_map_state.get(key)

    def contains(self, key) -> bool:
        return self._internal_map_state.contains(key)

    def items(self):
        return self._internal_map_state.items()

    def keys(self):
        return self._internal_map_state.keys()

    def values(self):
        return self._internal_map_state.values()

    def is_empty(self):
        return self._internal_map_state.is_empty()

    def clear(self):
        return self._internal_map_state.clear()


class SynchronousBroadcastRuntimeState(
    SynchronousReadOnlyBroadcastRuntimeState, InternalBroadcastState
):
    def __init__(self, name: str, internal_map_state: "InternalSynchronousMapRuntimeState"):
        super(SynchronousBroadcastRuntimeState, self).__init__(name, internal_map_state)

    def put(self, key, value):
        self._internal_map_state.put(key, value)

    def put_all(self, dict_value):
        self._internal_map_state.put_all(dict_value)

    def remove(self, key):
        self._internal_map_state.remove(key)

    def commit(self):
        self._internal_map_state.commit()

    def to_read_only_broadcast_state(self) -> "SynchronousReadOnlyBroadcastRuntimeState":
        return SynchronousReadOnlyBroadcastRuntimeState(self._name, self._internal_map_state)


class OperatorStateBackend(OperatorStateStore, ABC):

    @abstractmethod
    def commit(self):
        pass


class RemoteOperatorStateBackend(OperatorStateBackend):
    def __init__(
        self, state_handler, state_cache_size, map_state_read_cache_size, map_state_write_cache_size
    ):
        self._state_handler = state_handler
        self._state_cache_size = state_cache_size
        # NOTE: if user stores a state into a class member, that state actually won't be actually
        # evicted from memory (because its counter > 0)
        self._state_cache = LRUCache(state_cache_size, None)
        self._state_cache.set_on_evict(lambda _, state: state.commit())
        self._map_state_read_cache_size = map_state_read_cache_size
        self._map_state_write_cache_size = map_state_write_cache_size
        self._map_state_handler = CachingMapStateHandler(state_handler, map_state_read_cache_size)

    def get_broadcast_state(
        self, state_descriptor: MapStateDescriptor
    ) -> 'SynchronousBroadcastRuntimeState':
        state_name = state_descriptor.name
        map_coder = cast(MapCoder, from_type_info(state_descriptor.type_info))  # type: MapCoder
        key_coder = map_coder._key_coder
        value_coder = map_coder._value_coder

        if state_name in self._state_cache:
            self._validate_broadcast_state(state_name, key_coder, value_coder)
            return self._state_cache.get(state_name)

        state_proto = pb2_StateDescriptor()
        state_proto.state_name = state_name
        # Currently, MultimapKeysSideInput is used for BroadcastState
        state_key = beam_fn_api_pb2.StateKey(
            multimap_keys_side_input=beam_fn_api_pb2.StateKey.MultimapKeysSideInput(
                transform_id="",
                window=bytes(),
                side_input_id=base64.b64encode(state_proto.SerializeToString()),
            )
        )

        internal_map_state = InternalSynchronousMapRuntimeState(
            self._map_state_handler,
            state_key,
            key_coder,
            value_coder,
            self._map_state_write_cache_size,
        )

        broadcast_state = SynchronousBroadcastRuntimeState(
            state_descriptor.name, internal_map_state
        )
        self._state_cache.put(state_name, broadcast_state)
        return broadcast_state

    def commit(self):
        for state in self._state_cache:
            cast(SynchronousBroadcastRuntimeState, state).commit()

    def _validate_broadcast_state(self, name, key_coder, value_coder):
        if name in self._state_cache:
            state = cast(SynchronousBroadcastRuntimeState, self._state_cache.get(name))
            if (
                key_coder != state._internal_map_state._map_key_coder
                or value_coder != state._internal_map_state._map_value_coder
            ):
                raise Exception("State name corrupted: %s" % name)
